# -*- coding: utf-8 -*-
"""
A convinient interface to solving assignment problems with the Hungarian
algorithm (also known as Munkres or maximum linear-sum-assignment).

The core implementation of munkres in in scipy. Recent versions are written in
C, so their speed should be reflected here.

TODO:
   - [ ] Implement linear-time maximum weight matching approximation algorithm
     from this paper: https://web.eecs.umich.edu/~pettie/papers/ApproxMWM-JACM.pdf
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
# from distutils.version import LooseVersion
# import scipy
# SCIPY_GE_1_4 = LooseVersion(scipy.__version__) >= LooseVersion('1.4.0')


def mindist_assignment(vecs1, vecs2, p=2):
    """
    Finds minimum cost assignment between two sets of D dimensional vectors.

    Args:
        vecs1 (np.ndarray): NxD array of vectors representing items in vecs1
        vecs2 (np.ndarray): MxD array of vectors representing items in vecs2
        p (float): L-p norm to use. Default is 2 (aka Eucliedean)

    Returns:
        Tuple[list, float]: tuple containing assignments of rows in vecs1 to
            rows in vecs2, and the total distance between assigned pairs.

    Notes:
        Thin wrapper around mincost_assignment

    CommandLine:
        xdoctest -m ~/code/kwarray/kwarray/algo_assignment.py mindist_assignment

    CommandLine:
        xdoctest -m ~/code/kwarray/kwarray/algo_assignment.py mindist_assignment

    Example:
        >>> # xdoctest: +REQUIRES(module:scipy)
        >>> # Rows are detections in img1, cols are detections in img2
        >>> rng = np.random.RandomState(43)
        >>> vecs1 = rng.randint(0, 10, (5, 2))
        >>> vecs2 = rng.randint(0, 10, (7, 2))
        >>> ret = mindist_assignment(vecs1, vecs2)
        >>> print('Total error: {:.4f}'.format(ret[1]))
        Total error: 8.2361
        >>> print('Assignment: {}'.format(ret[0]))  # xdoc: +IGNORE_WANT
        Assignment: [(0, 0), (1, 3), (2, 5), (3, 2), (4, 6)]
    """
    from scipy.spatial import distance_matrix
    cost = distance_matrix(vecs1, vecs2, p=p)
    assignment, dist_tot = mincost_assignment(cost)
    return assignment, dist_tot


def mincost_assignment(cost):
    """
    Finds the minimum cost assignment based on a NxM cost matrix, subject to
    the constraint that each row can match at most one column and each column
    can match at most one row. Any pair with a cost of infinity will not be
    assigned.

    Args:
        cost (ndarray): NxM matrix, cost[i, j] is the cost to match i and j

    Returns:
        Tuple[list, float]: tuple containing a list of assignment of rows
            and columns, and the total cost of the assignment.

    CommandLine:
        xdoctest -m ~/code/kwarray/kwarray/algo_assignment.py mincost_assignment


    Example:
        >>> # xdoctest: +REQUIRES(module:scipy)
        >>> # Costs to match item i in set1 with item j in set2.
        >>> cost = np.array([
        >>>     [9, 2, 1, 9],
        >>>     [4, 1, 5, 5],
        >>>     [9, 9, 2, 4],
        >>> ])
        >>> ret = mincost_assignment(cost)
        >>> print('Assignment: {}'.format(ret[0]))
        >>> print('Total cost: {}'.format(ret[1]))
        Assignment: [(0, 2), (1, 1), (2, 3)]
        Total cost: 6

    Example:
        >>> # xdoctest: +REQUIRES(module:scipy)
        >>> cost = np.array([
        >>>     [0, 0, 0, 0],
        >>>     [4, 1, 5, -np.inf],
        >>>     [9, 9, np.inf, 4],
        >>>     [9, -2, np.inf, 4],
        >>> ])
        >>> ret = mincost_assignment(cost)
        >>> print('Assignment: {}'.format(ret[0]))
        >>> print('Total cost: {}'.format(ret[1]))
        Assignment: [(0, 2), (1, 3), (2, 0), (3, 1)]
        Total cost: -inf

    Example:
        >>> # xdoctest: +REQUIRES(module:scipy)
        >>> cost = np.array([
        >>>     [0, 0, 0, 0],
        >>>     [4, 1, 5, -3],
        >>>     [1, 9, np.inf, 0.1],
        >>>     [np.inf, np.inf, np.inf, 100],
        >>> ])
        >>> ret = mincost_assignment(cost)
        >>> print('Assignment: {}'.format(ret[0]))
        >>> print('Total cost: {}'.format(ret[1]))
        Assignment: [(0, 2), (1, 1), (2, 0), (3, 3)]
        Total cost: 102.0
    """
    from scipy.optimize import linear_sum_assignment
    n1, n2 = cost.shape
    n = max(n1, n2)
    # Embed the [n1 x n2] matrix in a padded (with inf) [n x n] matrix
    cost_matrix = np.full((n, n), fill_value=np.inf)
    cost_matrix[0:n1, 0:n2] = cost

    # Find an effective infinite value for infeasible assignments
    is_infinte = np.isinf(cost_matrix)
    is_finite = ~is_infinte
    is_positive = cost_matrix > 0
    is_negative = cost_matrix < 0
    # Note: in scipy 1.4 input costs may be infinte, should fix for this case
    # (also note, they don't allow a budgeted solution, so maybe we have to use
    # effective values)
    feasible_pos_vals = cost_matrix[(is_finite & is_positive)]
    feasible_neg_vals = cost_matrix[(is_finite & is_negative)]
    feasible_extent = feasible_pos_vals.sum() - feasible_neg_vals.sum()
    # Find a value that is approximately pos/neg infinity wrt this matrix
    approx_pos_inf = (n + feasible_extent) * 2
    approx_neg_inf = -approx_pos_inf
    # replace infinite values with effective infinite values
    cost_matrix[(is_infinte & is_positive)] = approx_pos_inf
    cost_matrix[(is_infinte & is_negative)] = approx_neg_inf

    # Solve munkres problem for minimum weight assignment
    indexes = list(zip(*linear_sum_assignment(cost_matrix)))
    # Return only the feasible assignments
    assignment = [(i, j) for (i, j) in indexes
                  if cost_matrix[i, j] < approx_pos_inf]
    # assert len(assignment) == min(cost.shape)
    cost_tot = sum([cost[i, j] for i, j in assignment])
    return assignment, cost_tot


def maxvalue_assignment(value):
    """
    Finds the maximum value assignment based on a NxM value matrix. Any pair
    with a non-positive value will not be assigned.

    Args:
        value (ndarray): NxM matrix, value[i, j] is the value of matching i and j

    Returns:
        Tuple[list, float]: tuple containing a list of assignment of rows
            and columns, and the total value of the assignment.

    CommandLine:
        xdoctest -m ~/code/kwarray/kwarray/algo_assignment.py maxvalue_assignment

    Example:
        >>> # xdoctest: +REQUIRES(module:scipy)
        >>> # Costs to match item i in set1 with item j in set2.
        >>> value = np.array([
        >>>     [9, 2, 1, 3],
        >>>     [4, 1, 5, 5],
        >>>     [9, 9, 2, 4],
        >>>     [-1, -1, -1, -1],
        >>> ])
        >>> ret = maxvalue_assignment(value)
        >>> # Note, depending on the scipy version the assignment might change
        >>> # but the value should always be the same.
        >>> print('Total value: {}'.format(ret[1]))
        Total value: 23.0
        >>> print('Assignment: {}'.format(ret[0]))  # xdoc: +IGNORE_WANT
        Assignment: [(0, 0), (1, 3), (2, 1)]

        >>> ret = maxvalue_assignment(np.array([[np.inf]]))
        >>> print('Assignment: {}'.format(ret[0]))
        >>> print('Total value: {}'.format(ret[1]))
        Assignment: [(0, 0)]
        Total value: inf

        >>> ret = maxvalue_assignment(np.array([[0]]))
        >>> print('Assignment: {}'.format(ret[0]))
        >>> print('Total value: {}'.format(ret[1]))
        Assignment: []
        Total value: 0
    """
    cost = (-value).astype(np.float)
    cost[value <= 0] = np.inf  # dont take anything with non-positive value
    assignment, cost_tot = mincost_assignment(cost)
    value_tot = -cost_tot
    return assignment, value_tot


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m kwarray.algo_assignment
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
"""
A faster-than-pandas pandas-like interface to column-major data, in the case
where the data only needs to be accessed by index.

For data where more complex ids are needed you must use pandas.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import ubelt as ub
import numpy as np
import copy

try:
    import pandas as pd
except Exception:
    pd = None


__version__ = '0.0.1'


class LocLight(object):
    def __init__(self, parent):
        self.parent = parent

    def __getitem__(self, index):
        return self.parent._getrow(index)


class DataFrameLight(ub.NiceRepr):
    r"""
    Implements a subset of the pandas.DataFrame API

    The API is restricted to facilitate speed tradeoffs

    Notes:
        Assumes underlying data is Dict[list|ndarray]. If the data is known
        to be a Dict[ndarray] use DataFrameArray instead, which has faster
        implementations for some operations.

    Notes:
        pandas.DataFrame is slow. DataFrameLight is faster.
        It is a tad more restrictive though.

    Example:
        >>> self = DataFrameLight({})
        >>> print('self = {!r}'.format(self))
        >>> self = DataFrameLight({'a': [0, 1, 2], 'b': [2, 3, 4]})
        >>> print('self = {!r}'.format(self))
        >>> item = self.iloc[0]
        >>> print('item = {!r}'.format(item))

    Benchmark:
        >>> # BENCHMARK
        >>> # xdoc: +REQUIRES(--bench)
        >>> from kwarray.dataframe_light import *  # NOQA
        >>> import ubelt as ub
        >>> NUM = 1000
        >>> print('NUM = {!r}'.format(NUM))
        >>> # to_dict conversions
        >>> print('==============')
        >>> print('====== to_dict conversions =====')
        >>> _keys = ['list', 'dict', 'series', 'split', 'records', 'index']
        >>> results = []
        >>> df = DataFrameLight._demodata(num=NUM).pandas()
        >>> ti = ub.Timerit(verbose=False, unit='ms')
        >>> for key in _keys:
        >>>     result = ti.reset(key).call(lambda: df.to_dict(orient=key))
        >>>     results.append((result.mean(), result.report()))
        >>> key = 'series+numpy'
        >>> result = ti.reset(key).call(lambda: {k: v.values for k, v in df.to_dict(orient='series').items()})
        >>> results.append((result.mean(), result.report()))
        >>> print('\n'.join([t[1] for t in sorted(results)]))
        >>> print('==============')
        >>> print('====== DFLight Conversions =======')
        >>> ti = ub.Timerit(verbose=True, unit='ms')
        >>> key = 'self.pandas'
        >>> self = DataFrameLight(df)
        >>> ti.reset(key).call(lambda: self.pandas())
        >>> key = 'light-from-pandas'
        >>> ti.reset(key).call(lambda: DataFrameLight(df))
        >>> key = 'light-from-dict'
        >>> ti.reset(key).call(lambda: DataFrameLight(self._data))
        >>> print('==============')
        >>> print('====== BENCHMARK: .LOC[] =======')
        >>> ti = ub.Timerit(num=20, bestof=4, verbose=True, unit='ms')
        >>> df_light = DataFrameLight._demodata(num=NUM)
        >>> # xdoctest: +REQUIRES(module:pandas)
        >>> df_heavy = df_light.pandas()
        >>> series_data = df_heavy.to_dict(orient='series')
        >>> list_data = df_heavy.to_dict(orient='list')
        >>> np_data = {k: v.values for k, v in df_heavy.to_dict(orient='series').items()}
        >>> for timer in ti.reset('DF-heavy.iloc'):
        >>>     with timer:
        >>>         for i in range(NUM):
        >>>             df_heavy.iloc[i]
        >>> for timer in ti.reset('DF-heavy.loc'):
        >>>     with timer:
        >>>         for i in range(NUM):
        >>>             df_heavy.iloc[i]
        >>> for timer in ti.reset('dict[SERIES].loc'):
        >>>     with timer:
        >>>         for i in range(NUM):
        >>>             {key: series_data[key].loc[i] for key in series_data.keys()}
        >>> for timer in ti.reset('dict[SERIES].iloc'):
        >>>     with timer:
        >>>         for i in range(NUM):
        >>>             {key: series_data[key].iloc[i] for key in series_data.keys()}
        >>> for timer in ti.reset('dict[SERIES][]'):
        >>>     with timer:
        >>>         for i in range(NUM):
        >>>             {key: series_data[key][i] for key in series_data.keys()}
        >>> for timer in ti.reset('dict[NDARRAY][]'):
        >>>     with timer:
        >>>         for i in range(NUM):
        >>>             {key: np_data[key][i] for key in np_data.keys()}
        >>> for timer in ti.reset('dict[list][]'):
        >>>     with timer:
        >>>         for i in range(NUM):
        >>>             {key: list_data[key][i] for key in np_data.keys()}
        >>> for timer in ti.reset('DF-Light.iloc/loc'):
        >>>     with timer:
        >>>         for i in range(NUM):
        >>>             df_light.iloc[i]
        >>> for timer in ti.reset('DF-Light._getrow'):
        >>>     with timer:
        >>>         for i in range(NUM):
        >>>             df_light._getrow(i)
        NUM = 1000
        ==============
        ====== to_dict conversions =====
        Timed best=0.022 ms, mean=0.022 ± 0.0 ms for series
        Timed best=0.059 ms, mean=0.059 ± 0.0 ms for series+numpy
        Timed best=0.315 ms, mean=0.315 ± 0.0 ms for list
        Timed best=0.895 ms, mean=0.895 ± 0.0 ms for dict
        Timed best=2.705 ms, mean=2.705 ± 0.0 ms for split
        Timed best=5.474 ms, mean=5.474 ± 0.0 ms for records
        Timed best=7.320 ms, mean=7.320 ± 0.0 ms for index
        ==============
        ====== DFLight Conversions =======
        Timed best=1.798 ms, mean=1.798 ± 0.0 ms for self.pandas
        Timed best=0.064 ms, mean=0.064 ± 0.0 ms for light-from-pandas
        Timed best=0.010 ms, mean=0.010 ± 0.0 ms for light-from-dict
        ==============
        ====== BENCHMARK: .LOC[] =======
        Timed best=101.365 ms, mean=101.564 ± 0.2 ms for DF-heavy.iloc
        Timed best=102.038 ms, mean=102.273 ± 0.2 ms for DF-heavy.loc
        Timed best=29.357 ms, mean=29.449 ± 0.1 ms for dict[SERIES].loc
        Timed best=21.701 ms, mean=22.014 ± 0.3 ms for dict[SERIES].iloc
        Timed best=11.469 ms, mean=11.566 ± 0.1 ms for dict[SERIES][]
        Timed best=0.807 ms, mean=0.826 ± 0.0 ms for dict[NDARRAY][]
        Timed best=0.478 ms, mean=0.492 ± 0.0 ms for dict[list][]
        Timed best=0.969 ms, mean=0.994 ± 0.0 ms for DF-Light.iloc/loc
        Timed best=0.760 ms, mean=0.776 ± 0.0 ms for DF-Light._getrow

    """
    def __init__(self, data=None, columns=None):
        if columns is not None:
            if data is None:
                data = ub.odict(zip(columns, [[] for _ in range(len(columns))]))
            else:
                data = ub.odict(zip(columns, data.T))

        self._raw = data
        self._data = None
        self._localizer = LocLight(self)
        self.__normalize__()

    @property
    def iloc(self):
        return self._localizer

    @property
    def values(self):
        data = self._getcols(self.columns)
        return data

    @property
    def loc(self):
        return self._localizer

    def __eq__(self, other):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:pandas)
            >>> self = DataFrameLight._demodata(num=7)
            >>> other = self.pandas()
            >>> assert np.all(self == other)
        """
        self_vals = self.values
        if isinstance(other, DataFrameLight):
            other_vals = other.values
        if pd is not None and isinstance(other, pd.DataFrame):
            other_vals = other.reindex(columns=self.columns).values
        else:
            other_vals = other
        return self_vals == other_vals

    def to_string(self, *args, **kwargs):
        return self.pandas().to_string(*args, **kwargs)

    def to_dict(self, orient='dict', into=dict):
        """
        Convert the data frame into a dictionary.

        Args:
            orient (str): Currently naitively suports orient in
                {'dict', 'list'}, otherwise we fallback to pandas conversion
                and call its to_dict method.

            into (type): type of dictionary to transform into

        Returns:
           dict

        Example:
            >>> from kwarray.dataframe_light import *  # NOQA
            >>> self = DataFrameLight._demodata(num=7)
            >>> print(self.to_dict(orient='dict'))
            >>> print(self.to_dict(orient='list'))
        """
        if orient == 'dict':
            out = into(self.iterrows())
        elif orient == 'list':
            import kwarray
            out = into((k, kwarray.ArrayAPI.tolist(v))
                        for k, v in self._data.items())
        else:
            out = self.pandas().to_dict(orient=orient, into=into)
        return out

    def pandas(self):
        """
        Convert back to pandas if you need the full API

        Example:
            >>> # xdoctest: +REQUIRES(module:pandas)
            >>> df_light = DataFrameLight._demodata(num=7)
            >>> df_heavy = df_light.pandas()
            >>> got = DataFrameLight(df_heavy)
            >>> assert got._data == df_light._data
        """
        if pd is None:
            raise Exception('Pandas is not available')
        return pd.DataFrame(self._data)

    def _pandas(self):
        """ Deprecated, use self.pandas instead """
        return self.pandas()

    @classmethod
    def _demodata(cls, num=7):
        """
        Example:
            >>> self = DataFrameLight._demodata(num=7)
            >>> print('self = {!r}'.format(self))
            >>> other = DataFrameLight._demodata(num=11)
            >>> print('other = {!r}'.format(other))
            >>> both = self.union(other)
            >>> print('both = {!r}'.format(both))
            >>> assert both is not self
            >>> assert other is not self
        """
        demodata = {
            'foo': [0] * num,
            'bar': [i % 3 for i in range(num)],
            'baz': [2.73] * num,
        }
        self = cls(demodata)
        return self

    def __nice__(self):
        return 'keys: {}, len={}'.format(list(self.keys()), len(self))

    def __len__(self):
        if self._data:
            key = next(iter(self.keys()))
            return len(self._data[key])
        else:
            return 0

    def __contains__(self, item):
        return item in self.keys()

    def __normalize__(self):
        """
        Try to convert input data to Dict[List]
        """
        if self._raw is None:
            self._data = {}
        elif isinstance(self._raw, dict):
            self._data = self._raw
            if __debug__:
                lens = []
                for d in self._data.values():
                    if not isinstance(d, (list, np.ndarray)):
                        raise TypeError(type(d))
                    lens.append(len(d))
                assert ub.allsame(lens)
        elif isinstance(self._raw, DataFrameLight):
            self._data = copy.copy(self._raw._data)
        elif pd is not None and isinstance(self._raw, pd.DataFrame):
            self._data = self._raw.to_dict(orient='list')
        else:
            raise TypeError('Unknown _raw type')

    @property
    def columns(self):
        return list(self.keys())

    def sort_values(self, key, inplace=False, ascending=True):
        sortx = np.argsort(self._getcol(key))
        if not ascending:
            sortx = sortx[::-1]
        return self.take(sortx, inplace=inplace)

    def keys(self):
        if self._data:
            for key in self._data.keys():
                yield key

    def _getrow(self, index):
        return {key: self._data[key][index] for key in self._data.keys()}

    def _getcol(self, key):
        return self._data[key]

    def _getcols(self, keys):
        num = len(self)
        col_data = [self._getcol(key) for key in keys]
        data = np.hstack([np.asarray(d).reshape(num, -1) for d in col_data])
        return data

    def get(self, key, default=None):
        """
        Get item for given key. Returns default value if not found.
        """
        return self[key] if key in self else default

    def clear(self):
        """
        Removes all rows inplace
        """
        if self._data:
            for key in self._data.keys():
                self._data[key].clear()

    def __getitem__(self, key):
        """
        Note:
            only handles the case where key is a single column name.

        Example:
            >>> df_light = DataFrameLight._demodata(num=7)
            >>> sub1 = df_light['bar']
            >>> # xdoctest: +REQUIRES(module:pandas)
            >>> df_heavy = df_light.pandas()
            >>> sub2 = df_heavy['bar']
            >>> assert np.all(sub1 == sub2)
        """
        return self._getcol(key)

    def __setitem__(self, key, value):
        """
        Note:
            only handles the case where key is a single column name. and value
            is an array of all the values to set.

        Example:
            >>> df_light = DataFrameLight._demodata(num=7)
            >>> value = [2] * len(df_light)
            >>> df_light['bar'] = value
            >>> # xdoctest: +REQUIRES(module:pandas)
            >>> df_heavy = df_light.pandas()
            >>> df_heavy['bar'] = value
            >>> assert np.all(df_light == df_heavy)
        """
        self._data[key] = value

    def compress(self, flags, inplace=False):
        """
        NOTE: NOT A PART OF THE PANDAS API
        """
        subset = self if inplace else self.__class__()
        for key in self._data.keys():
            subset._data[key] = list(ub.compress(self._data[key], flags))
        return subset

    def take(self, indices, inplace=False):
        """
        Return the elements in the given *positional* indices along an axis.

        Args:
            inplace (bool): NOT PART OF PANDAS API

        Notes:
            assumes axis=0

        Example:
            >>> df_light = DataFrameLight._demodata(num=7)
            >>> indices = [0, 2, 3]
            >>> sub1 = df_light.take(indices)
            >>> # xdoctest: +REQUIRES(module:pandas)
            >>> df_heavy = df_light.pandas()
            >>> sub2 = df_heavy.take(indices)
            >>> assert np.all(sub1 == sub2)
        """
        subset = self if inplace else self.__class__()
        if isinstance(indices, slice):
            for key in self._data.keys():
                subset._data[key] = self._data[key][indices]
        else:
            for key in self._data.keys():
                subset._data[key] = list(ub.take(self._data[key], indices))
        return subset

    def copy(self):
        other = copy.copy(self)
        other._data = other._data.copy()
        other._localizer = LocLight(other)
        return other

    def extend(self, other):
        """
        Extend ``self`` inplace using another dataframe array

        Args:
            other (DataFrameLight | dict[str, Sequence]):
                values to concat to end of this object

        NOTE:
            Not part of the pandas API

        Example:
            >>> self = DataFrameLight(columns=['foo', 'bar'])
            >>> other = {'foo': [0], 'bar': [1]}
            >>> self.extend(other)
            >>> assert len(self) == 1
        """
        try:
            _other_data = other._data
        except AttributeError:
            _other_data = other

        _self_data = self._data
        for key, vals1 in _self_data.items():
            vals2 = _other_data[key]
            try:
                vals1.extend(vals2)
            except AttributeError:
                if isinstance(vals1, np.ndarray):
                    _self_data[key] = np.hstack([vals1, vals2])
                else:
                    raise

    def union(self, *others):
        """
        NOTE:
            Note part of the pandas API
        """
        if isinstance(self, DataFrameLight):
            first = self
            rest = others
        else:
            if len(others) == 0:
                return DataFrameLight()
            first = others[0]
            rest = others[1:]

        both = first.copy()
        if not both.keys:
            for other in rest:
                if other.keys:
                    both.keys = copy.copy(other.keys)
                    break

        for other in rest:
            both.extend(other)
        return both

    @classmethod
    def concat(cls, others):
        return cls.union(*others)

    @classmethod
    def from_pandas(cls, df):
        _raw = {k: v.values for k, v in df.to_dict(orient='series').items()}
        return cls(_raw)

    @classmethod
    def from_dict(cls, records):
        record_iter = iter(records)
        columns = {}
        try:
            r = next(record_iter)
            for key, value in r.items():
                columns[key] = [value]
        except StopIteration:
            pass
        else:
            for r in record_iter:
                for key, value in r.items():
                    columns[key].append(value)
        self = cls(columns)
        return self

    def reset_index(self, drop=False):
        """ noop for compatability, the light version doesnt store an index """
        return self

    def groupby(self, by=None, *args, **kwargs):
        """
        Group rows by the value of a column. Unlike pandas this simply
        returns a zip object. To ensure compatiability call list on the
        result of groupby.

        Args:
            by (str): column name to group by
            *args: if specified, the dataframe is coerced to pandas
            *kwargs: if specified, the dataframe is coerced to pandas

        Example:
            >>> df_light = DataFrameLight._demodata(num=7)
            >>> res1 = list(df_light.groupby('bar'))
            >>> # xdoctest: +REQUIRES(module:pandas)
            >>> df_heavy = df_light.pandas()
            >>> res2 = list(df_heavy.groupby('bar'))
            >>> assert len(res1) == len(res2)
            >>> assert all([np.all(a[1] == b[1]) for a, b in zip(res1, res2)])

        Ignore:
            >>> self = DataFrameLight._demodata(num=1000)
            >>> args = ['cx']
            >>> self['cx'] = (np.random.rand(len(self)) * 10).astype(np.int)
            >>> # As expected, our custom restricted implementation is faster
            >>> # than pandas
            >>> ub.Timerit(100).call(lambda: dict(list(self.pandas().groupby('cx')))).print()
            >>> ub.Timerit(100).call(lambda: dict(self.groupby('cx'))).print()
        """
        if len(args) == 0 and len(kwargs) == 0:
            # In this special case we can be fast
            import kwarray
            unique, groupxs = kwarray.group_indices(self[by])
            groups = [self.take(idxs) for idxs in groupxs]
            return zip(unique, groups)
        else:
            # otherwise we need to use the slow method
            return self.pandas().groupby(by=by)

    def rename(self, mapper=None, columns=None, axis=None, inplace=False):
        """
        Rename the columns (index renaming is not supported)

        Example:
            >>> df_light = DataFrameLight._demodata(num=7)
            >>> mapper = {'foo': 'fi'}
            >>> res1 = df_light.rename(columns=mapper)
            >>> res3 = df_light.rename(mapper, axis=1)
            >>> # xdoctest: +REQUIRES(module:pandas)
            >>> df_heavy = df_light.pandas()
            >>> res2 = df_heavy.rename(columns=mapper)
            >>> res4 = df_heavy.rename(mapper, axis=1)
            >>> assert np.all(res1 == res2)
            >>> assert np.all(res3 == res2)
            >>> assert np.all(res3 == res4)
        """
        if columns is not None:
            if axis is not None:
                raise TypeError("Cannot specify both 'axis' and any of 'index' or 'columns'")
        else:
            if axis != 1:
                raise NotImplementedError('only axis=1 is supported')
            columns = mapper

        if not inplace:
            self = self.copy()
        for old, new in columns.items():
            if old in self._data:
                self._data[new] = self._data.pop(old)
        return self

    def iterrows(self):
        """
        Iterate over rows as (index, Dict) pairs.

        Yields:
            Tuple[int, Dict]: the index and a dictionary representing a row

        Example:
            >>> from kwarray.dataframe_light import *  # NOQA
            >>> self = DataFrameLight._demodata(num=3)
            >>> print(ub.repr2(list(self.iterrows())))
            [
                (0, {'bar': 0, 'baz': 2.73, 'foo': 0}),
                (1, {'bar': 1, 'baz': 2.73, 'foo': 0}),
                (2, {'bar': 2, 'baz': 2.73, 'foo': 0}),
            ]

        Benchmark:
            >>> # xdoc: +REQUIRES(--bench)
            >>> from kwarray.dataframe_light import *  # NOQA
            >>> import ubelt as ub
            >>> df_light = DataFrameLight._demodata(num=1000)
            >>> df_heavy = df_light.pandas()
            >>> ti = ub.Timerit(21, bestof=3, verbose=2, unit='ms')
            >>> ti.reset('light').call(lambda: list(df_light.iterrows()))
            >>> ti.reset('heavy').call(lambda: list(df_heavy.iterrows()))
            >>> # xdoctest: +IGNORE_WANT
            Timed light for: 21 loops, best of 3
                time per loop: best=0.834 ms, mean=0.850 ± 0.0 ms
            Timed heavy for: 21 loops, best of 3
                time per loop: best=45.007 ms, mean=45.633 ± 0.5 ms
        """
        for idx in range(len(self)):
            row = self._getrow(idx)
            yield idx, row


class DataFrameArray(DataFrameLight):
    """
    DataFrameLight assumes the backend is a Dict[list]
    DataFrameArray assumes the backend is a Dict[ndarray]

    Take and compress are much faster, but extend and union are slower
    """

    def __normalize__(self):
        """
        Try to convert input data to Dict[ndarray]
        """
        if self._raw is None:
            self._data = {}
        elif isinstance(self._raw, dict):
            self._data = self._raw
            if __debug__:
                lens = []
                for d in self._data.values():
                    if not isinstance(d, (list, np.ndarray)):
                        raise TypeError(type(d))
                    lens.append(len(d))
                assert ub.allsame(lens), (
                    'lens are not all same {} for columns {}'.format(
                        lens,
                        list(self._data.keys()))
                )
        elif isinstance(self._raw, DataFrameLight):
            self._data = copy.copy(self._raw._data)
        elif pd is not None and isinstance(self._raw, pd.DataFrame):
            self._data = {k: v.values for k, v in self._raw.to_dict(orient='series').items()}
        else:
            raise TypeError('Unknown _raw type')
        # self._data = ub.map_vals(np.asarray, self._data)  # does this break anything?

    def extend(self, other):
        for key in self._data.keys():
            vals1 = self._data[key]
            vals2 = other._data[key]
            self._data[key] = np.hstack([vals1, vals2])

    def compress(self, flags, inplace=False):
        subset = self if inplace else self.__class__()
        for key in self._data.keys():
            subset._data[key] = self._data[key][flags]
        return subset

    def take(self, indices, inplace=False):
        subset = self if inplace else self.__class__()
        for key in self._data.keys():
            subset._data[key] = self._data[key][indices]
        return subset

    # def min(self, axis=None):
    #     return self._extreme(func=np.minimum, axis=axis)

    # def max(self, axis=None):
    #     """
    #     Example:
    #         >>> self = DataFrameArray._demodata(num=7)
    #         >>> func = np.maximum
    #     """
    #     return self._extreme(func=np.maximum, axis=axis)

    # def _extreme(self, func, axis=None):
    #     if axis is None:
    #         raise NotImplementedError
    #     if axis == 0:
    #         raise NotImplementedError
    #     elif axis == 1:
    #         newdata = nh.util.iter_reduce_ufunc(func, (self[key] for key in self.keys()))
    #         newobj = self.__class__(newdata, self._keys)
    #     else:
    #         raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Defines data structures for efficient repeated sampling of specific
distributions (e.g. Normal, Uniform, Binomial) with specific parameters.

Inspired by ~/code/imgaug/imgaug/parameters.py

Similar Libraries:
    * https://docs.pymc.io/api/distributions.html
    * https://github.com/phobson/paramnormal

TODO:
    - [ ] change sample shape to just a single num.
    - [ ] Some Distributions will output vectors. Maybe we could just postpend the dimensions?

"""
from __future__ import print_function, division, absolute_import, unicode_literals
import numpy as np
import ubelt as ub
import functools
from kwarray.util_random import ensure_rng


def _issubclass2(child, parent):
    """
    Uses string comparisons to avoid ipython reload errors.
    Much less robust though.
    """
    # String comparison
    if child.__name__ == parent.__name__:
        if child.__module__ == parent.__module__:
            return True
    # Recurse through classes of obj
    return any(_issubclass2(base, parent) for base in child.__bases__)


def _isinstance2(obj, cls):
    """
    obj = self
    cls = distributions.Distribution

    child = obj.__class__
    parent = cls
    """
    return isinstance(obj, cls)
    # try:
    #     return _issubclass2(obj.__class__, cls)
    # except Exception:
    #     return False


class Parameterized(ub.NiceRepr):
    """
    Keeps track of all registered params and classes with registered params
    """
    def __init__(self):
        self._called_init = True
        self._params = ub.odict()
        self._children = ub.odict()

    def __setattr__(self, key, value):
        if not getattr(self, '_called_init', key == '_called_init'):
            raise Exception(
                'Need to call super().__init__() before setting '
                'attribute "{}" of Parameterized classes'.format(key))

        if not key.startswith('_'):
            if key in self._children or _isinstance2(value, Parameterized):
                self._children[key] = value
        super(Parameterized, self).__setattr__(key, value)

    def _setparam(self, key, value):
        setattr(self, key, value)
        self._params[key] = value

    def _setchild(self, key, value):
        assert isinstance(key, str)
        self._children[key] = value

    def children(self):
        for key, value in self._children.items():
            yield key, value

    def seed(self, rng=None):
        rng = ensure_rng(rng)
        for key, child in self.children():
            if _isinstance2(child, Parameterized):
                child.seed(rng)

    def parameters(self):
        """
        Returns parameters in this object and its children
        """
        for key, value in self._params.items():
            yield key, value
        for prefix, child in self.children():
            for subkey, value in child.parameters():
                key = prefix + '.' + subkey
                yield key,  value

    def _body_str():
        pass

    def idstr(self, nl=None, thresh=80):
        """
        Example:
            >>> self = TruncNormal()
            >>> self.idstr()
            >>> #
            >>> #
            >>> class Dummy(Distribution):
            >>>     def __init__(self):
            >>>         super(Dummy, self).__init__()
            >>>         self._setparam('a', 3)
            >>>         self.b = Normal()
            >>>         self.c = Uniform()
            >>> self = Dummy()
            >>> print(self.idstr())
            >>> #
            >>> class Tail5(Distribution):
            >>>     def __init__(self):
            >>>         super(Tail5, self).__init__()
            >>>         self._setparam('a_parameter', 3)
            >>>         for i in range(5):
            >>>             self._setparam(chr(i + 97), i)
            >>> #
            >>> class Tail6(Distribution):
            >>>     def __init__(self):
            >>>         super(Tail6, self).__init__()
            >>>         for i in range(9):
            >>>             self._setparam(chr(i + 97) + '_parameter', i)
            >>> #
            >>> class Dummy2(Distribution):
            >>>     def __init__(self):
            >>>         super(Dummy2, self).__init__()
            >>>         self._setparam('x', 3)
            >>>         self._setparam('y', 3)
            >>>         self.d = Dummy()
            >>>         self.f = Tail6()
            >>>         self.y = Tail5()
            >>> self = Dummy2()
            >>> print(self.idstr())
            >>> print(ub.repr2(self.json_id()))
        """
        classname = self.__class__.__name__
        self_part = ['{}={}'.format(key, ub.repr2(value, precision=2, si=True, nl=0))
                     for key, value in self._params.items()]
        child_part = ['{}={}'.format(key, child.idstr(nl, thresh=thresh - 2))
                      for key, child in self.children()]

        body, nl = self._make_body(self_part, child_part, nl, thresh - len(classname) - 2)
        if nl:
            body = ub.indent(body, '  ')
        return '{}({})'.format(classname, body.rstrip(' '))

    def json_id(self):
        children = ub.odict([(key, child.json_id())
                             for key, child in self.children()])
        params = ub.odict([
            (key, value.tolist() if isinstance(value, np.ndarray) else value)
            for key, value in self._params.items()])
        return ub.dict_union(ub.odict([('__class__', self.__class__.__name__)]),
                             params,
                             children)

    def _make_body(self, self_part, child_part, nl=None, thresh=80):
        parts = self_part + child_part
        if len(parts) == 0:
            body = ''
        else:
            sep = ', '
            oneline_body = sep.join(parts)
            if nl is None:
                nl = len(oneline_body) > thresh or len(child_part) > 0
            if nl:
                sep = ',\n'
                body = '\n' + sep.join(parts) + '\n'
            else:
                body = oneline_body
        return body, nl

    def __nice__(self):
        try:
            params = ub.odict(self.parameters())
            return ub.repr2(params, nl=0, precision=2, si=True, nobr=True,
                            explicit=True)
        except Exception:
            return '?'


class ParameterizedList(Parameterized):
    """

    Example:
        >>> from kwarray import distributions as stoch
        >>> self1 = stoch.ParameterizedList([
        >>>     stoch.Normal(),
        >>>     stoch.Uniform(),
        >>> ])
        >>> print(self1.idstr())
        >>> self = stoch.ParameterizedList([stoch.ParameterizedList([
        >>>     stoch.Normal(),
        >>>     stoch.Uniform(),
        >>>     self1,
        >>> ])])
        >>> print(self.idstr())
        >>> print(self.idstr(0))

    """
    def __init__(self, items):
        super(ParameterizedList, self).__init__()
        self._items = []
        for t in items:
            self.append(t)

    def _setparam(self, key, value):
        raise NotImplementedError('lists cannot have params')

    def append(self, item):
        key = str(len(self._items))
        self._items.append(item)
        self._setchild(key, item)

    def __iter__(self):
        return iter(self._items)

    def __getitem__(self, key):
        if isinstance(key, (int, slice)):
            return self._items[key]
        else:
            return self._children[key]

    def idstr(self, nl=None, thresh=80):
        assert len(self._items) == len(self._children)
        assert len(self._params) == 0
        child_part = ['{}'.format(child.idstr(nl, thresh=thresh - 2))
                      for key, child in self.children()]
        body, nl = self._make_body([], child_part, nl, thresh - 2)
        if nl:
            body = ub.indent(body, '  ')
        return '[{}]'.format(body.rstrip(' '))


class _BinOpMixin(object):
    def __add__(self, other):
        return Composed(np.add, [self, other])

    def __pow__(self, other):
        return Composed(np.power, [self, other])

    def __mul__(self, other):
        return Composed(np.multiply, [self, other])

    def __sub__(self, other):
        return Composed(np.subtract, [self, other])

    def __truediv__(self, other):
        return Composed(np.divide, [self, other])

    def __matmul__(self, other):
        raise NotImplementedError

    def __floordiv__(self, other):
        raise NotImplementedError

    def __mod__(self, other):
        raise NotImplementedError

    def __divmod__(self, other):
        raise NotImplementedError

    def __lshift__(self, other):
        raise NotImplementedError

    def __rshift__(self, other):
        raise NotImplementedError

    def __and__(self, other):
        raise NotImplementedError

    def __xor__(self, other):
        raise NotImplementedError

    def __or__(self, other):
        raise NotImplementedError

    def __neg__(self):
        return Composed(np.negative, [self])

    def __pos__(self):
        return Composed(np.positive, [self])

    def __invert__(self):
        raise NotImplementedError

    def __abs__(self):
        return Composed(np.abs, [self])

    def __round__(self, ndigits=None):
        return Composed(np.round, [self, ndigits])

    def __trunc__(self):
        return Composed(np.trunc, [self])

    def __floor__(self):
        return Composed(np.floor, [self])

    def __ceil__(self):
        return Composed(np.ceil, [self])

    # Numpy mixins

    # def __getattr__(self, attr):
    #     # Handle application of general numpy functions
    #     operation = getattr(np, attr)
    #     return Composed(operation, [self])

    def int(self):
        def _cast_int(x):
            if isinstance(x, np.ndarray):
                return x.astype(np.int)
            else:
                return int(x)
        return Composed(_cast_int, [self])

    def round(self, ndigits=None):
        return Composed(np.round, [self, ndigits])

    def clip(self, a_min=None, a_max=None):
        return Composed(np.clip, [self, a_min, a_max])

    def log(self):
        return Composed(np.log, [self])

    def log10(self):
        return Composed(np.log10, [self])

    def exp(self):
        return Composed(np.exp, [self])

    def sqrt(self):
        return Composed(np.sqrt, [self])

    def abs(self):
        return Composed(np.abs, [self])


class _RBinOpMixin(_BinOpMixin):
    """
    https://docs.python.org/3/reference/datamodel.html
    """
    def __radd__(self, other):
        return Composed(np.add, [other, self])

    def __rpow__(self, other):
        return Composed(np.power, [other, self])

    def __rmul__(self, other):
        return Composed(np.multiply, [other, self])

    def __rsub__(self, other):
        return Composed(np.subtract, [other, self])

    def __rtruediv__(self, other):
        return Composed(np.divide, [other, self])

    def __rmatmul__(self, other):
        raise NotImplementedError

    def __rfloordiv__(self, other):
        raise NotImplementedError

    def __rmod__(self, other):
        raise NotImplementedError

    def __rdivmod__(self, other):
        raise NotImplementedError

    def __rlshift__(self, other):
        raise NotImplementedError

    def __rrshift__(self, other):
        raise NotImplementedError

    def __rand__(self, other):
        raise NotImplementedError

    def __rxor__(self, other):
        raise NotImplementedError

    def __ror__(self, other):
        raise NotImplementedError


class Distribution(Parameterized, _RBinOpMixin):
    """
    Base class for all distributions
    """
    def __init__(self, rng=None):
        super(Distribution, self).__init__()
        self.rng = ensure_rng(rng)

    def __call__(self, *shape):
        return self.sample(*shape)

    def seed(self, rng=None):
        super(Distribution, self).seed(rng)
        self.rng = ensure_rng(rng)

    def sample(self, *shape):
        raise NotImplementedError('overwrite me')

    @classmethod
    def cast(cls, arg):
        if isinstance(arg, cls):
            return arg
        elif isinstance(arg, Distribution):
            # allow distribution substitution
            return arg
        elif isinstance(arg, (list, tuple)):
            return cls(*arg)
        elif isinstance(arg, dict):
            return cls(**arg)
        else:
            raise CastError('cannot cast {} as {}'.format(arg, cls))

    @classmethod
    def seeded(cls, rng=0):
        return Seeded(rng, cls)

    def _show(self, n, bins=None, ax=None, color=None, label=None):
        """ plot samples monte-carlo style """
        if ax is None:
            import kwplot
            kwplot.autompl()
            from matplotlib import pyplot as plt
            ax = plt.gca()
        data = self.sample(n)
        ax.hist(data, bins=bins, color=color, label=label)


class Mixture(Distribution):
    """
    Creates a mixture model of multiple distributions

    Args:
        pdfs (list): list of distributions
        weights (list): corresponding weights of each distribution
        rng (np.random.RandomState): seed random number generator

    Example:
        >>> # In this examle we create a bimodal mixture of normals
        >>> pdfs = [Normal(mean=10, std=2), Normal(18, 2)]
        >>> self = Mixture(pdfs)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.figure(fnum=1, doclf=True)
        >>> self._show(500, bins=25)
    """
    def __init__(self, pdfs, weights=None, rng=None):
        super(Mixture, self).__init__(rng=rng)
        self.pdfs = pdfs
        self._setparam('pdfs', pdfs)
        self._setparam('weights', weights)
        idxs = np.arange(len(pdfs))
        self._idx_choice = Categorical(idxs, weights, rng=rng)

    def sample(self, *shape):
        """
        Sampling from a mixture of k distributions with weights w_k is
        equivalent to picking a distribution with probability w_k, and then
        sampling from the picked distribution.
        """
        # Choose which distributions are picked for each sample
        idxs = self._idx_choice.sample(*shape)
        idx_to_nsamples = ub.dict_hist(idxs.ravel())
        out = np.zeros(*shape)
        for idx, n in idx_to_nsamples.items():
            # Sample the from the distribution we picked
            mask = (idx == idxs)
            subsample = self.pdfs[idx].sample(n)
            out[mask] = subsample
        return out


class Composed(Distribution):
    """
    A distribution generated by composing different base distributions.

    Note this is not the same as mixing distributions!

    Attributes:
        self.operation (Function) : operation (add / sub / mult / div) to
            perform on operands
        self.operands (Sequence) : arguments passed to operation

    Example:
        >>> # In this examle you can see that the sum of two Normal random
        >>> # variables is also normal
        >>> operands = [Normal(mean=10, std=2), Normal(15, 2)]
        >>> operation = np.add
        >>> self = Composed(operation, operands)
        >>> data = self.sample(5)
        >>> print(ub.repr2(list(data), nl=0, precision=5))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.figure(fnum=1, doclf=True)
        >>> self._show(1000, bins=100)

    Example:
        >>> # Binary operations result in composed distributions
        >>> # We can make a (bounded) exponential distribution using a uniform
        >>> X = Uniform(.001, 7)
        >>> lam = .7
        >>> e = np.exp(1)
        >>> self = lam * e ** (-lam * X)
        >>> data = self.sample(5)
        >>> print(ub.repr2(list(data), nl=0, precision=5))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.figure(fnum=1, doclf=True)
        >>> self._show(5000, bins=100)
    """
    def __init__(self, operation, operands):
        super(Composed, self).__init__()
        self._setparam('operation', operation)
        self._setparam('operands', operands)
        # self.operation = operation
        # self.operands = operands

    def sample(self, *shape):
        # resolved_args = [arg.sample(*shape) for arg in self.operands]
        resolved_args = [_trysample(arg, shape) for arg in self.operands]
        return self.operation(*resolved_args)


def _trysample(arg, shape):
    """ samples if arg is a distribution, otherwise returns arg """
    try:
        return arg.sample(*shape)
    except Exception:
        return arg


class CastError(ValueError):
    pass


class Uniform(Distribution):
    """
    Defaults to a uniform distribution between 0 and 1

    Example:
        >>> self = Uniform(rng=0)
        >>> self.sample()
        0.548813...
        >>> float(self.sample(1))
        0.7151...

    Benchmark:
        >>> import ubelt as ub
        >>> self = Uniform()
        >>> for timer in ub.Timerit(100, bestof=10):
        >>>     with timer:
        >>>         [self() for _ in range(100)]
        >>> for timer in ub.Timerit(100, bestof=10):
        >>>     with timer:
        >>>         self(100)
    """
    def __init__(self, low=0, high=1, rng=None):
        super(Uniform, self).__init__(rng=rng)
        self._setparam('low', low)
        self._setparam('high', high)

    def sample(self, *shape):
        return self.rng.rand(*shape) * (self.high - self.low) + self.low

    @classmethod
    def cast(cls, arg):
        try:
            self = super(Uniform, cls).cast(arg)
        except CastError:
            if isinstance(arg, (int, float)):
                self = cls(low=0, high=arg)
            else:
                raise
        return self


class Exponential(Distribution):
    """

    Example:
        >>> self = Exponential(rng=0)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.figure(fnum=1, doclf=True)
        >>> self._show(500, bins=25)
    """
    def __init__(self, scale=1, rng=None):
        super(Exponential, self).__init__(rng=rng)
        self._setparam('scale', scale)

    def sample(self, *shape):
        return self.rng.exponential(self.scale, *shape)


class Constant(Distribution):
    """
    Example:
        >>> self = Constant(42, rng=0)
        >>> self.sample()
        42
        >>> self.sample(3)
        array([42, 42, 42])
    """
    def __init__(self, value=0, rng=None):
        super(Constant, self).__init__(rng=rng)
        self._setparam('value', value)

    def sample(self, *shape):
        if shape:
            return np.full(shape, fill_value=self.value)
        else:
            return self.value


class DiscreteUniform(Distribution):
    """
    Max is exclusive

    Example:
        >>> self = DiscreteUniform.cast(4)
        >>> self.sample(100)
    """
    def __init__(self, min=0, max=1, rng=None):
        super(DiscreteUniform, self).__init__(rng=rng)
        self._setparam('min', min)
        self._setparam('max', max)

    def sample(self, *shape):
        if len(shape) == 0:
            shape = None
        idx = self.rng.randint(self.min, self.max, size=shape)
        return idx

    @classmethod
    def cast(cls, arg):
        try:
            self = super(DiscreteUniform, cls).cast(arg)
        except CastError:
            if isinstance(arg, (int, float)):
                self = cls(max=arg)
            else:
                raise
        return self


class Normal(Distribution):
    """
        >>> self = Normal(mean=100, rng=0)
        >>> self.sample()
        >>> self.sample(100)
    """
    def __init__(self, mean=0, std=1, rng=None):
        super(Normal, self).__init__(rng=rng)
        self._setparam('mean', mean)
        self._setparam('std', std)

    def sample(self, *shape):
        return self.rng.randn(*shape) * self.std + self.mean


class Bernoulli(Distribution):
    """
    self = Normal()
    self.sample()
    self.sample(1)
    """
    def __init__(self, p=.5, rng=None):
        super(Bernoulli, self).__init__(rng=rng)
        assert p >= 0
        assert p <= 1
        self._setparam('p', p)

    def sample(self, *shape):
        return self.rng.rand(*shape) < self.p

    @classmethod
    def cast(cls, arg):
        try:
            self = super(Bernoulli, cls).cast(arg)
        except CastError:
            if isinstance(arg, (int, float)):
                self = cls(p=arg)
            else:
                raise
        return self


class Binomial(Distribution):
    """
    self = Normal()
    self.sample()
    self.sample(1)
    """
    def __init__(self, n=1, p=.5, rng=None):
        super(Binomial, self).__init__(rng=rng)
        self._setparam('p', p)
        self._setparam('n', n)

    def sample(self, *shape):
        return self.rng.rand(*shape) > self.p


class Categorical(Distribution):
    """
    Example:
        >>> categories = [3, 5, 1]
        >>> weights = [.05, .5, .45]
        >>> self = Categorical(categories, weights, rng=0)
        >>> self.sample()
        5
        >>> list(self.sample(2))
        [1, 1]
        >>> self.sample(2, 3)
        array([[5, 5, 1],
               [5, 1, 1]])
    """
    def __init__(self, categories, weights=None, rng=None):
        super(Categorical, self).__init__(rng=rng)
        self._setparam('categories', np.array(categories))
        self._setparam('weights', weights)
        self._idxs = np.arange(len(self.categories))

    def sample(self, *shape):
        if len(shape) == 0:
            shape = None
        idxs = self.rng.choice(self._idxs, size=shape, p=self.weights)
        # if isinstance(self.categories, np.ndarray):
        return self.categories[idxs]
        # return self.rng.choice(self.categories, size=shape, p=self.weights)


class NonlinearUniform(Distribution):
    """
    Weighted sample between two points depending on some nonlinearity

    TODO:
        could refactor part of this into a PowerLaw distribution

    Args:
        nonlinearity (func or str): needs to be a function that maps
             the range 0-1 to the range 0-1

    Example:
        >>> self = NonlinearUniform(0, 100, np.sqrt, rng=0)
        >>> print(ub.repr2(list(self.sample(2)), precision=2, nl=0))
        [74.08, 84.57]
        >>> print(ub.repr2(self.sample(2, 3), precision=2, nl=1))
        np.array([[77.64, 73.82, 65.09],
                  [80.37, 66.15, 94.43]], dtype=np.float64)

    Ignore:
        bins = np.linspace(0, 100, 100)
        NonlinearUniform(0, 100, None, rng=0)._show(10000, bins=bins)
        NonlinearUniform(0, 100, 'sqrt', rng=0)._show(10000, bins=bins)
        NonlinearUniform(0, 100, 'squared', rng=0)._show(10000, bins=bins)
        NonlinearUniform(0, 100, 'log', rng=0)._show(10000, bins=bins)
        NonlinearUniform(0, 100, 'exp', rng=0)._show(10000, bins=bins)
        NonlinearUniform(0, 100, 0.25, rng=0)._show(10000, bins=bins)
        NonlinearUniform(0, 100, 4.0, rng=0)._show(10000, bins=bins)
        NonlinearUniform(0, 100, 2.0, rng=0)._show(10000, bins=bins)
    """
    def __init__(self, min, max, nonlinearity=None, reverse=False, rng=None):
        super(NonlinearUniform, self).__init__(rng=rng)
        self._setparam('min', min)
        self._setparam('max', max)
        if nonlinearity == 'linear' or nonlinearity is None:
            nonlinearity = None
        elif nonlinearity == 'exp':
            nonlinearity = lambda x: (np.exp(x) - 1) / (np.exp(1) - 1)  # NOQA
        elif nonlinearity == 'log':
            nonlinearity = lambda x: np.log(x + 1) / np.log(2)  # NOQA
        elif nonlinearity == 'squared':
            nonlinearity = lambda x: np.power(x, 2)  # NOQA
        elif nonlinearity == 'sqrt':
            nonlinearity = np.sqrt
        elif isinstance(nonlinearity, float):
            # Use a powerlaw
            param = nonlinearity
            nonlinearity = lambda x: np.power(x, param)  # NOQA
        elif not callable(nonlinearity):
            raise KeyError(nonlinearity)
        self._setparam('nonlinearity', nonlinearity)
        self._setparam('reverse', reverse)

    def sample(self, *shape):
        if len(shape) == 0:
            base = self.rng.rand()
        else:
            base = self.rng.rand(*shape)
        if self.reverse:
            base = 1 - base
        if self.nonlinearity is not None:
            base = self.nonlinearity(base)
        return base * (self.max - self.min) + self.min


class CategoryUniform(Distribution):
    """
    Discrete Uniform over a list of categories
    """
    def __init__(self, categories=[None], rng=None):
        super(CategoryUniform, self).__init__(rng=rng)
        self._setparam('categories', np.array(categories))
        self._num = len(self.categories)

    def sample(self, *shape):
        if len(shape) == 0:
            shape = None
        idx = self.rng.randint(0, self._num, size=shape)
        return self.categories[idx]


class TruncNormal(Distribution):
    """
    A truncated normal distribution.

    A normal distribution, but bounded by low and high values. Note this is
    much different from just using a clipped normal.

    Args:
        mean (float): mean of the distribution
        std (float): standard deviation of the distribution
        low (float): lower bound
        high (float): upper bound
        rng (np.random.RandomState):

    Example:
        >>> self = TruncNormal(rng=0)
        >>> self()  # output of this changes before/after scipy version 1.5
        ...0.1226...

    Example:
        >>> low = -np.pi / 16
        >>> high = np.pi / 16
        >>> std = np.pi / 8
        >>> self = TruncNormal(low=low, high=high, std=std, rng=0)
        >>> shape = (3, 3)
        >>> data = self(*shape)
        >>> print(ub.repr2(data, precision=5))
        np.array([[ 0.01841,  0.0817 ,  0.0388 ],
                  [ 0.01692, -0.0288 ,  0.05517],
                  [-0.02354,  0.15134,  0.18098]], dtype=np.float64)
    """
    def __init__(self, mean=0, std=1, low=-np.inf, high=np.inf, rng=None):
        super(TruncNormal, self).__init__(rng=rng)
        self._setparam('mean', mean)
        self._setparam('std', std)
        self._setparam('low', low)
        self._setparam('high', high)
        self._update_internals()

    def _update_internals(self):
        from scipy.stats import truncnorm
        # Convert high and low values to be wrt the standard normal range
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.truncnorm.html
        self.a = (self.low - self.mean) / self.std
        self.b = (self.high - self.mean) / self.std
        self.rv = truncnorm(a=self.a, b=self.b, loc=self.mean, scale=self.std)

    def sample(self, *shape):
        arr = self.rv.rvs(size=shape, random_state=self.rng)
        return arr


class PDF(Distribution):
    """
    BROKEN?

    Similar to Catigorical, but interpolates to approximate a continuous random
    variable.

    Returns a value x with probability p.

    References:
        http://www.nehalemlabs.net/prototype/blog/2013/12/16/how-to-do-inverse-transformation-sampling-in-scipy-and-numpy/

    Args:
        x (list or tuple): domain in which this PDF is defined
        p (list): probability sample for each domain sample

    Example:
        >>> x = np.linspace(800, 4500)
        >>> p = np.log10(x)
        >>> p = x ** 2
        >>> self = PDF(x, p)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.figure(fnum=1, doclf=True)
        >>> self._show(5000, bins=50)
    """
    def __init__(self, x, p, rng=None):
        import scipy.interpolate as interpolate
        super(PDF, self).__init__(rng=rng)

        if isinstance(x, tuple):
            self.edge_x = np.linspace(x[0], x[1], len(p) + 1)
        else:
            # Need to convert point samples to edges
            self.edge_x = interpolate.interp1d(np.linspace(0, 1, len(x)), x)(np.linspace(0, 1, len(x) + 1))
            # FIXME: might have a bug causing probability to be shifted slightly

        # Normalize probabilities
        p = p / p.sum()

        # Calculate CDF and interpolate edges
        self.edge_cdf = np.hstack([[0], np.cumsum(p)])

        # FIXME: This is not actually the inverse. Like, at all.
        # In fact its the inverse of the inverse.
        self.inv_cdf = interpolate.interp1d(self.edge_cdf, self.edge_x)

    def sample(self, *shape):
        # Right idea, but wrong implementation
        r = np.random.rand(*shape)
        return self.inv_cdf(r)


class Seeded(object):
    """
    Helper for grabbing pre-seeded distributions
    """
    def __init__(self, rng=None, cls=None):
        self.rng = rng
        self.cls = cls

    def __call__(self, *args, **kwargs):
        return self.cls(*args, rng=self.rng, **kwargs)

    def __getattr__(self, key):
        return functools.partial(globals()[key], rng=self.rng)


def _test_distributions():
    from kwarray import distributions as stoch
    cls_list = []
    for cls in vars(stoch).values():
        if isinstance(cls, type):
            if issubclass(cls, stoch.Distribution):
                if cls is not stoch.Distribution:
                    cls_list.append(cls)

    for cls in cls_list:
        rng = np.random.RandomState(0)
        inst = cls(rng=rng)
        scalar_sample = inst.sample()
        vector_sample = inst.sample(1)
        print('inst = {!r}'.format(inst))
        print('scalar_sample = {!r}'.format(scalar_sample))
        print('vector_sample = {!r}'.format(vector_sample))
        assert not ub.iterable(scalar_sample)
        assert ub.iterable(vector_sample)
        assert vector_sample.shape == (1,)


"""
General Dev References:
    https://stackoverflow.com/questions/21100716/fast-arbitrary-distribution-random-sampling
    https://codereview.stackexchange.com/questions/196286/inverse-transform-sampling
"""



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import ubelt as ub
import numpy as np
import itertools as it

try:
    import torch
except ImportError:
    torch = None


class SlidingWindow(ub.NiceRepr):
    """
    Slide a window of a certain shape over an array with a larger shape.

    This can be used for iterating over a grid of sub-regions of 2d-images,
    3d-volumes, or any n-dimensional array.

    Yields slices of shape `window` that can be used to index into an array
    with shape `shape` via numpy / torch fancy indexing. This allows for fast
    fast iteration over subregions of a larger image. Because we generate a
    grid-basis using only shapes, the larger image does not need to be in
    memory as long as its width/height/depth/etc...

    Args:
        shape (Tuple[int, ...]): shape of source array to slide across.

        window (Tuple[int, ...]): shape of window that will be slid over the
            larger image.

        overlap (float, default=0): a number between 0 and 1 indicating the
            fraction of overlap that parts will have. Specifying this is
            mutually exclusive with `stride`.  Must be `0 <= overlap < 1`.

        stride (int, default=None): the number of cells (pixels) moved on each
            step of the window. Mutually exclusive with overlap.

        keepbound (bool, default=False): if True, a non-uniform stride will be
            taken to ensure that the right / bottom of the image is returned as
            a slice if needed. Such a slice will not obey the overlap
            constraints.  (Defaults to False)

        allow_overshoot (bool, default=False): if False, we will raise an
            error if the window doesn't slide perfectly over the input shape.

    Attributes:
        basis_shape - shape of the grid corresponding to the number of strides
            the sliding window will take.
        basis_slices - slices that will be taken in every dimension

    Yields:
        Tuple[slice, ...]: slices used for numpy indexing, the number of slices
            in the tuple

    Notes:
        For each dimension, we generate a basis (which defines a grid), and we
        slide over that basis.

    Example:
        >>> from kwarray.util_slider import *  # NOQA
        >>> shape = (10, 10)
        >>> window = (5, 5)
        >>> self = SlidingWindow(shape, window)
        >>> for i, index in enumerate(self):
        >>>     print('i={}, index={}'.format(i, index))
        i=0, index=(slice(0, 5, None), slice(0, 5, None))
        i=1, index=(slice(0, 5, None), slice(5, 10, None))
        i=2, index=(slice(5, 10, None), slice(0, 5, None))
        i=3, index=(slice(5, 10, None), slice(5, 10, None))

    Example:
        >>> from kwarray.util_slider import *  # NOQA
        >>> shape = (16, 16)
        >>> window = (4, 4)
        >>> self = SlidingWindow(shape, window, overlap=(.5, .25))
        >>> print('self.stride = {!r}'.format(self.stride))
        self.stride = [2, 3]
        >>> list(ub.chunks(self.grid, 5))
        [[(0, 0), (0, 1), (0, 2), (0, 3), (0, 4)],
         [(1, 0), (1, 1), (1, 2), (1, 3), (1, 4)],
         [(2, 0), (2, 1), (2, 2), (2, 3), (2, 4)],
         [(3, 0), (3, 1), (3, 2), (3, 3), (3, 4)],
         [(4, 0), (4, 1), (4, 2), (4, 3), (4, 4)],
         [(5, 0), (5, 1), (5, 2), (5, 3), (5, 4)],
         [(6, 0), (6, 1), (6, 2), (6, 3), (6, 4)]]
    """
    def __init__(self, shape, window, overlap=None, stride=None,
                 keepbound=False, allow_overshoot=False):
        if overlap is None and stride is None:
            overlap = 0

        assert len(window) == len(shape), (
            'incompatible dims: {} {}'.format(len(window),
                                              len(shape)))
        # assert all(d <= D for d, D in zip(window, shape)), (
        #         'window must be smaller than shape')

        stride, overlap = self._compute_stride(overlap, stride, shape,
                                               window)

        # print('stride = {!r}'.format(stride))
        if not all(stride):
            raise ValueError(
                'Step must be positive everywhere. Got={}'.format(stride))

        stide_kw = [dict(margin=d, stop=D, step=s, keepbound=keepbound,
                         check=not keepbound and not allow_overshoot)
                      for d, D, s in zip(window, shape, stride)]

        undershot_shape = []
        overshoots = []
        for kw in stide_kw:
            final_pos = (kw['stop'] - kw['margin'])
            n_steps = final_pos // kw['step']
            overshoot = final_pos % kw['step']
            undershot_shape.append(n_steps + 1)
            overshoots.append(overshoot)

        self._final_step = overshoots

        if not allow_overshoot and any(overshoots):
            raise ValueError('overshoot={} stide_kw={}'.format(overshoots,
                                                               stide_kw))

        # make a slice generator for each dimension
        self.stride = stride
        self.overlap = overlap

        self.window = window
        self.input_shape = shape

        # The undershot basis shape, only contains indices that correspond
        # perfectly to the input. It may crop a bit of the ends.  If this is
        # equal to basis_shape, then the self perfectly fits the input.
        self.undershot_shape = undershot_shape

        # NOTE: if we have overshot, then basis shape will not perfectly
        # align to the original image. This shape will be a bit bigger.
        self.basis_slices = [list(_slices1d(**kw)) for kw in stide_kw]
        self.basis_shape = [len(b) for b in self.basis_slices]
        self.n_total = np.prod(self.basis_shape)

    def __nice__(self):
        return 'bshape={}, shape={}, window={}, stride={}'.format(
            tuple(self.basis_shape),
            tuple(self.input_shape),
            self.window,
            tuple(self.stride)
        )

    def _compute_stride(self, overlap, stride, shape, window):
        """
        Ensures that stride hasoverlap the correct shape.  If stride is not
        provided, compute stride from desired overlap.
        """
        if isinstance(stride, np.ndarray):
            stride = tuple(stride)
        if isinstance(overlap, np.ndarray):
            overlap = tuple(overlap)

        if not (overlap is None) ^ (stride is None):
            raise ValueError('specify overlap({}) XOR stride ({})'.format(
                overlap, stride))
        if stride is None:
            if not isinstance(overlap, (list, tuple)):
                overlap = [overlap] * len(window)
            if any(frac < 0 or frac >= 1 for frac in overlap):
                raise ValueError((
                    'part overlap was {}, but fractional overlaps must be '
                    'in the range [0, 1)').format(overlap))
            stride = [int(round(d - d * frac))
                      for frac, d in zip(overlap, window)]
        else:
            if not isinstance(stride, (list, tuple)):
                stride = [stride] * len(window)
        # Recompute fractional overlap after integer stride is computed
        overlap = [(d - s) / d for s, d in zip(stride, window)]
        assert len(stride) == len(shape), 'incompatible dims'
        return stride, overlap

    def __len__(self):
        return self.n_total

    def _iter_basis_frac(self):
        for slices in self:
            frac = [sl.start / D for sl, D in zip(slices, self.source.shape)]
            yield frac

    def __iter__(self):
        for slices in it.product(*self.basis_slices):
            yield slices

    def __getitem__(self, index):
        """
        Get a specific item by its flat (raveled) index

        Example:
            >>> from kwarray.util_slider import *  # NOQA
            >>> window = (10, 10)
            >>> shape = (20, 20)
            >>> self = SlidingWindow(shape, window, stride=5)
            >>> itered_items = list(self)
            >>> assert len(itered_items) == len(self)
            >>> indexed_items = [self[i] for i in range(len(self))]
            >>> assert itered_items[0] == self[0]
            >>> assert itered_items[-1] == self[-1]
            >>> assert itered_items == indexed_items
        """
        if index < 0:
            index = len(self) + index
        # Find the nd location in the grid
        basis_idx = np.unravel_index(index, self.basis_shape)
        # Take the slice for each of the n dimensions
        slices = tuple([bdim[i]
                        for bdim, i in zip(self.basis_slices, basis_idx)])
        return slices

    @property
    def grid(self):
        """
        Generate indices into the "basis" slice for each dimension.
        This enumerates the nd indices of the grid.

        Yields:
            Tuple[int, ...]
        """
        # Generates basis for "sliding window" slices to break a large image
        # into smaller pieces. Use it.product to slide across the coordinates.
        basis_indices = map(range, self.basis_shape)
        for basis_idxs in it.product(*basis_indices):
            yield basis_idxs

    @property
    def slices(self):
        """
        Generate slices for each window (equivalent to iter(self))

        Example:
            >>> shape = (220, 220)
            >>> window = (10, 10)
            >>> self = SlidingWindow(shape, window, stride=5)
            >>> list(self)[41:45]
            [(slice(0, 10, None), slice(205, 215, None)),
             (slice(0, 10, None), slice(210, 220, None)),
             (slice(5, 15, None), slice(0, 10, None)),
             (slice(5, 15, None), slice(5, 15, None))]
            >>> print('self.overlap = {!r}'.format(self.overlap))
            self.overlap = [0.5, 0.5]
        """
        return iter(self)

    @property
    def centers(self):
        """
        Generate centers of each window

        Yields:
            Tuple[float, ...]: the center coordinate of the slice

        Example:
            >>> shape = (4, 4)
            >>> window = (3, 3)
            >>> self = SlidingWindow(shape, window, stride=1)
            >>> list(zip(self.centers, self.slices))
            [((1.0, 1.0), (slice(0, 3, None), slice(0, 3, None))),
             ((1.0, 2.0), (slice(0, 3, None), slice(1, 4, None))),
             ((2.0, 1.0), (slice(1, 4, None), slice(0, 3, None))),
             ((2.0, 2.0), (slice(1, 4, None), slice(1, 4, None)))]
            >>> shape = (3, 3)
            >>> window = (2, 2)
            >>> self = SlidingWindow(shape, window, stride=1)
            >>> list(zip(self.centers, self.slices))
            [((0.5, 0.5), (slice(0, 2, None), slice(0, 2, None))),
             ((0.5, 1.5), (slice(0, 2, None), slice(1, 3, None))),
             ((1.5, 0.5), (slice(1, 3, None), slice(0, 2, None))),
             ((1.5, 1.5), (slice(1, 3, None), slice(1, 3, None)))]
        """
        for slices in self:
            center = tuple(sl_.start + (sl_.stop - sl_.start - 1) / 2
                           for sl_ in slices)
            yield center


class Stitcher(ub.NiceRepr):
    """
    Stitches multiple possibly overlapping slices into a larger array.

    This is used to invert the SlidingWindow.  For semenatic segmentation the
    patches are probability chips. Overlapping chips are averaged together.

    Args:
        shape (tuple): dimensions of the large image that will be created from
            the smaller pixels or patches.

    TODO:
        - [ ] Look at the old "add_fast" code in the netharn version and see if
              it is worth porting. This code is kept in the dev folder in
              ../dev/_dev_slider.py

    Example:
        >>> from kwarray.util_slider import *  # NOQA
        >>> import sys
        >>> # Build a high resolution image and slice it into chips
        >>> highres = np.random.rand(5, 200, 200).astype(np.float32)
        >>> target_shape = (1, 50, 50)
        >>> slider = SlidingWindow(highres.shape, target_shape, overlap=(0, .5, .5))
        >>> # Show how Sticher can be used to reconstruct the original image
        >>> stitcher = Stitcher(slider.input_shape)
        >>> for sl in list(slider):
        ...     chip = highres[sl]
        ...     stitcher.add(sl, chip)
        >>> assert stitcher.weights.max() == 4, 'some parts should be processed 4 times'
        >>> recon = stitcher.finalize()
    """
    def __init__(stitcher, shape, device='numpy'):
        stitcher.shape = shape
        stitcher.device = device
        if device == 'numpy':
            stitcher.sums = np.zeros(shape, dtype=np.float32)
            stitcher.weights = np.zeros(shape, dtype=np.float32)

            stitcher.sumview = stitcher.sums.ravel()
            stitcher.weightview = stitcher.weights.ravel()
        else:
            stitcher.sums = torch.zeros(shape, device=device)
            stitcher.weights = torch.zeros(shape, device=device)

            stitcher.sumview = stitcher.sums.view(-1)
            stitcher.weightview = stitcher.weights.view(-1)

    def __nice__(stitcher):
        return str(stitcher.sums.shape)

    def add(stitcher, indices, patch, weight=None):
        """
        Incorporate a new (possibly overlapping) patch or pixel using a
        weighted sum.

        Args:
            indices (slice or tuple): typically a Tuple[slice] of pixels or a
                single pixel, but this can be any numpy fancy index.
            patch (ndarray): data to patch into the bigger image.
            weight (float or ndarray): weight of this patch (default to 1.0)
        """
        if weight is None:
            stitcher.sums[indices] += patch
            stitcher.weights[indices] += 1.0
        else:
            stitcher.sums[indices] += (patch * weight)
            stitcher.weights[indices] += weight

    def average(stitcher):
        """
        Averages out contributions from overlapping adds using weighted average

        Returns:
            out: ndarray: the stitched image
        """
        out = stitcher.sums / stitcher.weights
        return out

    def finalize(stitcher, indices=None):
        """
        Averages out contributions from overlapping adds

        Args:
            indices (None | slice | tuple): if None, finalize the entire
                block, otherwise only finalize a subregion.

        Returns:
            final: ndarray: the stitched image
        """
        if indices is None:
            final = stitcher.sums / stitcher.weights
        else:
            final = stitcher.sums[indices] / stitcher.weights[indices]

        # if stitcher.device != 'numpy':
        #     final = final.cpu().numpy()
        # final = np.nan_to_num(final)
        return final


def _slices1d(margin, stop, step=None, start=0, keepbound=False, check=True):
    """
    Helper to generates slices in a single dimension.

    Args:
        start (int): starting point (in most cases set this to 0)

        margin (int): the length of the slice (window)

        stop (int): the length of the image dimension

        step (int): the length of each step / distance between slices

        keepbound (bool): if True, a non-uniform step will be taken to ensure
            that the right / bottom of the image is returned as a slice if
            needed. Such a slice will not obey the overlap constraints.
            (Defaults to False)

        check (bool): if True an error will be raised if the window does not
            cover the entire extent from start to stop, even if keepbound is
            True.

    Yields:
        slice : slice in one dimension of size (margin)

    Example:
        >>> stop, margin, step = 2000, 360, 360
        >>> keepbound = True
        >>> strides = list(_slices1d(margin, stop, step, keepbound, check=False))
        >>> assert all([(s.stop - s.start) == margin for s in strides])

    Example:
        >>> stop, margin, step = 200, 46, 7
        >>> keepbound = True
        >>> strides = list(_slices1d(margin, stop, step, keepbound=False, check=True))
        >>> starts = np.array([s.start for s in strides])
        >>> stops = np.array([s.stop for s in strides])
        >>> widths = stops - starts
        >>> assert np.all(np.diff(starts) == step)
        >>> assert np.all(widths == margin)

    Example:
        >>> import pytest
        >>> stop, margin, step = 200, 36, 7
        >>> with pytest.raises(ValueError):
        ...     list(_slices1d(margin, stop, step))
    """
    if step is None:
        step = margin

    if check:
        # see how far off the end we would fall if we didnt check bounds
        perfect_final_pos = (stop - start - margin)
        overshoot = perfect_final_pos % step
        if overshoot > 0:
            raise ValueError(
                ('margin={} and step={} overshoot endpoint={} '
                 'by {} units when starting from={}').format(
                     margin, step, stop, overshoot, start))
    pos = start
    # probably could be more efficient with numpy here
    while True:
        endpos = pos + margin
        yield slice(pos, endpos)
        # Stop once we reached the end
        if endpos == stop:
            break
        pos += step
        if pos + margin > stop:
            if keepbound:
                # Ensure the boundary is always used even if steps
                # would overshoot Could do some other strategy here
                pos = stop - margin
            else:
                break



# -*- coding: utf-8 -*-
"""
Currently just defines "stats_dict", which is a nice way to gather multiple
numeric statistics (e.g. max, min, median, mode, arithmetic-mean,
geometric-mean, standard-deviation, etc...) about data in an array.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import collections
import numpy as np
import ubelt as ub


try:
    import torch
except Exception:
    torch = None


def stats_dict(inputs, axis=None, nan=False, sum=False, extreme=True,
               n_extreme=False, median=False, shape=True, size=False):
    """
    Describe statistics about an input array

    Args:
        inputs (ArrayLike): set of values to get statistics of
        axis (int): if ``inputs`` is ndarray then this specifies the axis
        nan (bool): report number of nan items
        sum (bool): report sum of values
        extreme (bool): report min and max values
        n_extreme (bool): report extreme value frequencies
        median (bool): report median
        size (bool): report array size
        shape (bool): report array shape

    Returns:
        collections.OrderedDict: stats: dictionary of common numpy statistics
            (min, max, mean, std, nMin, nMax, shape)

    SeeAlso:
        scipy.stats.describe

    Example:
        >>> # xdoctest: +IGNORE_WHITESPACE
        >>> from kwarray.util_averages import *  # NOQA
        >>> axis = 0
        >>> rng = np.random.RandomState(0)
        >>> inputs = rng.rand(10, 2).astype(np.float32)
        >>> stats = stats_dict(inputs, axis=axis, nan=False, median=True)
        >>> import ubelt as ub  # NOQA
        >>> result = str(ub.repr2(stats, nl=1, precision=4, with_dtype=True))
        >>> print(result)
        {
            'mean': np.array([ 0.5206,  0.6425], dtype=np.float32),
            'std': np.array([ 0.2854,  0.2517], dtype=np.float32),
            'min': np.array([ 0.0202,  0.0871], dtype=np.float32),
            'max': np.array([ 0.9637,  0.9256], dtype=np.float32),
            'med': np.array([0.5584, 0.6805], dtype=np.float32),
            'shape': (10, 2),
        }

    Example:
        >>> # xdoctest: +IGNORE_WHITESPACE
        >>> axis = 0
        >>> rng = np.random.RandomState(0)
        >>> inputs = rng.randint(0, 42, size=100).astype(np.float32)
        >>> inputs[4] = np.nan
        >>> stats = stats_dict(inputs, axis=axis, nan=True)
        >>> import ubelt as ub  # NOQA
        >>> result = str(ub.repr2(stats, nl=0, precision=1, strkeys=True))
        >>> print(result)
        {mean: 20.0, std: 13.2, min: 0.0, max: 41.0, num_nan: 1, shape: (100,)}
    """
    stats = collections.OrderedDict([])

    # Ensure input is in numpy format
    if isinstance(inputs, np.ndarray):
        nparr = inputs
    elif isinstance(inputs, list):
        nparr = np.array(inputs)
    elif torch is not None and isinstance(inputs, torch.Tensor):
        nparr = inputs.data.cpu().numpy()
    else:
        nparr = np.array(list(inputs))
    # Check to make sure stats are feasible
    if len(nparr) == 0:
        stats['empty_list'] = True
        if size:
            stats['size'] = 0
    else:
        if nan:
            min_val = np.nanmin(nparr, axis=axis)
            max_val = np.nanmax(nparr, axis=axis)
            mean_ = np.nanmean(nparr, axis=axis)
            std_  = np.nanstd(nparr, axis=axis)
        else:
            min_val = nparr.min(axis=axis)
            max_val = nparr.max(axis=axis)
            mean_ = nparr.mean(axis=axis)
            std_  = nparr.std(axis=axis)
        # number of entries with min/max val
        nMin = np.sum(nparr == min_val, axis=axis)
        nMax = np.sum(nparr == max_val, axis=axis)

        # Notes:
        # the first central moment is 0
        # the first raw moment is the mean
        # the second central moment is the variance
        # the third central moment is the skweness * var ** 3
        # the fourth central moment is the kurtosis * var ** 4
        # the third standardized moment is the skweness
        # the fourth standardized moment is the kurtosis

        if True:
            stats['mean'] = np.float32(mean_)
            stats['std'] = np.float32(std_)
        if extreme:
            stats['min'] = np.float32(min_val)
            stats['max'] = np.float32(max_val)
        if n_extreme:
            stats['nMin'] = np.int32(nMin)
            stats['nMax'] = np.int32(nMax)
        if median:
            stats['med'] = np.nanmedian(nparr, axis=axis)
        if nan:
            stats['num_nan'] = np.isnan(nparr).sum()
        if sum:
            sumfunc = np.nansum if nan else np.sum
            stats['sum'] = sumfunc(nparr, axis=axis)
        if size:
            stats['size'] = nparr.size
        if shape:
            stats['shape'] = nparr.shape
    return stats


def _gmean(a, axis=0, dtype=None, clobber=False):
    """
    Compute the geometric mean along the specified axis.

    Modification of the scikit-learn method to be more memory efficient

    Example
        >>> rng = np.random.RandomState(0)
        >>> C, H, W = 8, 32, 32
        >>> axis = 0
        >>> a = [rng.rand(C, H, W).astype(np.float16),
        >>>      rng.rand(C, H, W).astype(np.float16)]

    """
    if isinstance(a, np.ndarray):
        if clobber:
            # NOTE: we reuse (a), we clobber the input array!
            log_a = np.log(a, out=a)
        else:
            log_a = np.log(a)
    else:
        if dtype is None:
            # if not an ndarray object attempt to convert it
            log_a = np.log(np.array(a, dtype=dtype))
        else:
            # Must change the default dtype allowing array type
            # Note: that this will use memory, but there isn't anything we can
            # do here.
            if isinstance(a, np.ma.MaskedArray):
                a_ = np.ma.asarray(a, dtype=dtype)
            else:
                a_ = np.asarray(a, dtype=dtype)
            # We can reuse `a_` because it was a temp var
            log_a = np.log(a_, out=a_)

    # attempt to reuse memory when computing mean
    mem = log_a[axis]
    mean_log_a = log_a.mean(axis=axis, out=mem)

    # And reuse memory again when computing the final result
    result = np.exp(mean_log_a, out=mean_log_a)
    return result


class RunningStats(ub.NiceRepr):
    """
    Dynamically records per-element array statistics and can summarized them
    per-element, across channels, or globally.

    TODO:
        - [ ] This may need a few API tweaks and good documentation

    Example:
        >>> run = RunningStats()
        >>> ch1 = np.array([[0, 1], [3, 4]])
        >>> ch2 = np.zeros((2, 2))
        >>> img = np.dstack([ch1, ch2])
        >>> run.update(np.dstack([ch1, ch2]))
        >>> run.update(np.dstack([ch1 + 1, ch2]))
        >>> run.update(np.dstack([ch1 + 2, ch2]))
        >>> # No marginalization
        >>> print('current-ave = ' + ub.repr2(run.summarize(axis=ub.NoParam), nl=2, precision=3))
        >>> # Average over channels (keeps spatial dims separate)
        >>> print('chann-ave(k=1) = ' + ub.repr2(run.summarize(axis=0), nl=2, precision=3))
        >>> print('chann-ave(k=0) = ' + ub.repr2(run.summarize(axis=0, keepdims=0), nl=2, precision=3))
        >>> # Average over spatial dims (keeps channels separate)
        >>> print('spatial-ave(k=1) = ' + ub.repr2(run.summarize(axis=(1, 2)), nl=2, precision=3))
        >>> print('spatial-ave(k=0) = ' + ub.repr2(run.summarize(axis=(1, 2), keepdims=0), nl=2, precision=3))
        >>> # Average over all dims
        >>> print('alldim-ave(k=1) = ' + ub.repr2(run.summarize(axis=None), nl=2, precision=3))
        >>> print('alldim-ave(k=0) = ' + ub.repr2(run.summarize(axis=None, keepdims=0), nl=2, precision=3))
        """

    def __init__(run):
        run.raw_max = -np.inf
        run.raw_min = np.inf
        run.raw_total = 0
        run.raw_squares = 0
        run.n = 0

    def __nice__(self):
        return '{}'.format(self.shape)

    @property
    def shape(run):
        try:
            return run.raw_total.shape
        except Exception:
            return None

    def update(run, data):
        """
        Updates statistics across all data dimensions on a per-element basis
        """
        run.n += 1
        run.raw_max = np.maximum(run.raw_max, data)
        run.raw_min = np.minimum(run.raw_min, data)
        run.raw_total += data
        run.raw_squares += data ** 2

    def _sumsq_std(run, total, squares, n):
        """
        Sum of squares method to compute standard deviation
        """
        numer = (n * squares - total ** 2)
        denom = (n * (n - 1.0))
        std = np.sqrt(numer / denom)
        return std

    def summarize(run, axis=None, keepdims=True):
        """
        Compute summary statistics across a one or more dimension

        Args:
            axis (int | List[int] | None | ub.NoParam):
                axis or axes to summarize over,
                if None, all axes are summarized.
                if ub.NoParam, no axes are summarized the current result is
                    returned.

            keepdims (bool, default=True):
                if False removes the dimensions that are summarized over

        Returns:
            Dict: containing minimum, maximum, mean, std, etc..
        """
        if axis is ub.NoParam:
            total = run.raw_total
            squares = run.raw_squares
            maxi = run.raw_max
            mini = run.raw_min
            n = run.n
            info = ub.odict([
                ('n', n),
                ('max', maxi),
                ('min', mini),
                ('total', total),
                ('squares', squares),
                ('mean', total / n),
                ('std', run._sumsq_std(total, squares, n)),
            ])
            return info
        else:
            if run.n <= 0:
                raise RuntimeError('No statistics have been accumulated')
            total   = run.raw_total.sum(axis=axis, keepdims=keepdims)
            squares = run.raw_squares.sum(axis=axis, keepdims=keepdims)
            maxi    = run.raw_max.max(axis=axis, keepdims=keepdims)
            mini    = run.raw_min.min(axis=axis, keepdims=keepdims)
            if not hasattr(run.raw_total, 'shape'):
                n = run.n
            elif axis is None:
                n = run.n * np.prod(run.raw_total.shape)
            else:
                n = run.n * np.prod(np.take(run.raw_total.shape, axis))

            info = ub.odict([
                ('n', n),
                ('max', maxi),
                ('min', mini),
                ('total', total),
                ('squares', squares),
                ('mean', total / n),
                ('std', run._sumsq_std(total, squares, n)),
            ])
            return info

    def current(run):
        """
        Returns current staticis on a per-element basis
        (not summarized over any axis)

        TODO:
            - [X] I want this method and summarize to be unified somehow.
                I don't know how to paramatarize it because axis=None usually
                means summarize over everything, and I need to way to encode,
                summarize over nothing but the "sequence" dimension (which was
                given incrementally by the update function), which is what
                this function does.
        """
        info = run.summarize(axis=ub.NoParam)
        return info

if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/kwarray/kwarray/util_averages.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
Torch specific extensions
"""
import numpy as np
try:
    import torch
except Exception:
    torch = None


def one_hot_embedding(labels, num_classes, dim=1):
    """
    Embedding labels to one-hot form.

    Args:
        labels: (LongTensor) class labels, sized [N,].
        num_classes: (int) number of classes.
        dim (int): dimension which will be created, if negative

    Returns:
        Tensor: encoded labels, sized [N,#classes].

    References:
        https://discuss.pytorch.org/t/convert-int-into-one-hot-format/507/4

    Example:
        >>> # each element in target has to have 0 <= value < C
        >>> # xdoctest: +REQUIRES(module:torch)
        >>> labels = torch.LongTensor([0, 0, 1, 4, 2, 3])
        >>> num_classes = max(labels) + 1
        >>> t = one_hot_embedding(labels, num_classes)
        >>> assert all(row[y] == 1 for row, y in zip(t.numpy(), labels.numpy()))
        >>> import ubelt as ub
        >>> print(ub.repr2(t.numpy().tolist()))
        [
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0],
        ]
        >>> t2 = one_hot_embedding(labels.numpy(), num_classes)
        >>> assert np.all(t2 == t.numpy())
        >>> if torch.cuda.is_available():
        >>>     t3 = one_hot_embedding(labels.to(0), num_classes)
        >>>     assert np.all(t3.cpu().numpy() == t.numpy())

    Example:
        >>> # xdoctest: +REQUIRES(module:torch)
        >>> nC = num_classes = 3
        >>> labels = (torch.rand(10, 11, 12) * nC).long()
        >>> assert one_hot_embedding(labels, nC, dim=0).shape == (3, 10, 11, 12)
        >>> assert one_hot_embedding(labels, nC, dim=1).shape == (10, 3, 11, 12)
        >>> assert one_hot_embedding(labels, nC, dim=2).shape == (10, 11, 3, 12)
        >>> assert one_hot_embedding(labels, nC, dim=3).shape == (10, 11, 12, 3)
        >>> labels = (torch.rand(10, 11) * nC).long()
        >>> assert one_hot_embedding(labels, nC, dim=0).shape == (3, 10, 11)
        >>> assert one_hot_embedding(labels, nC, dim=1).shape == (10, 3, 11)
        >>> labels = (torch.rand(10) * nC).long()
        >>> assert one_hot_embedding(labels, nC, dim=0).shape == (3, 10)
        >>> assert one_hot_embedding(labels, nC, dim=1).shape == (10, 3)
    """
    if torch is not None and torch.is_tensor(labels):
        in_dims = labels.ndimension()
        if dim < 0:
            dim = in_dims - dim + 1
        if dim == 1 and in_dims == 1:
            # normal case where everything is already flat
            y = torch.eye(int(num_classes), device=labels.device)
            y_onehot = y[labels]
        else:
            # non-flat case (note that this would handle the normal case, but
            # why do extra work?)
            y = torch.eye(int(num_classes), device=labels.device)
            flat_y_onehot = y[labels.view(-1)]
            y_onehot = flat_y_onehot.view(*list(labels.shape) + [num_classes])
            if dim != in_dims:
                dim_order = list(range(in_dims))
                dim_order.insert(dim, in_dims)
                y_onehot = y_onehot.permute(*dim_order)
    else:
        if dim < 0:
            dim = labels.ndim - dim + 1
        flag = (dim != 1 or labels.ndim == 2)
        if flag:
            orig_shape = labels.shape
            labels = labels.reshape(-1)
            # raise NotImplementedError('not implemented for this case')
        y = np.eye(int(num_classes))
        y_onehot = y[labels]
        if flag:
            new_shape =  list(orig_shape) + [num_classes]
            y_onehot = y_onehot.reshape(*new_shape)
            old_axes = list(range(len(orig_shape)))
            new_axes = old_axes
            new_axes.insert(dim, len(orig_shape))
            y_onehot = y_onehot.transpose(*new_axes)
    return y_onehot


def one_hot_lookup(data, indices):
    """
    Return value of a particular column for each row in data.

    Each item in labels corresonds to a row in ``data``. Returns the index
    specified at each row.

    Args:
        data (ArrayLike): N x C float array of values
        indices (ArrayLike): N integer array between 0 and C.
            This is an column index for each row in ``data``.

    Returns:
        ArrayLike: the selected probability for each row

    Notes:
        This is functionally equivalent to
        ``[row[c] for row, c in zip(data, indices)]`` except that it is
        works with pure matrix operations.

    TODO:
        - [ ] Allow the user to specify which dimension indices should be
              zipped over. By default it should be dim=0

        - [ ] Allow the user to specify which dimension indices should select
              from. By default it should be dim=1.

    Example:
        >>> from kwarray.util_torch import *  # NOQA
        >>> data = np.array([
        >>>     [0, 1, 2],
        >>>     [3, 4, 5],
        >>>     [6, 7, 8],
        >>>     [9, 10, 11],
        >>> ])
        >>> indices = np.array([0, 1, 2, 1])
        >>> res = one_hot_lookup(data, indices)
        >>> print('res = {!r}'.format(res))
        res = array([ 0,  4,  8, 10])
        >>> alt = np.array([row[c] for row, c in zip(data, indices)])
        >>> assert np.all(alt == res)


    Example:
        >>> # xdoctest: +REQUIRES(module:torch)
        >>> import torch
        >>> data = torch.from_numpy(np.array([
        >>>     [0, 1, 2],
        >>>     [3, 4, 5],
        >>>     [6, 7, 8],
        >>>     [9, 10, 11],
        >>> ]))
        >>> indices = torch.from_numpy(np.array([0, 1, 2, 1])).long()
        >>> res = one_hot_lookup(data, indices)
        >>> print('res = {!r}'.format(res))
        res = tensor([ 0,  4,  8, 10]...)
        >>> alt = torch.LongTensor([row[c] for row, c in zip(data, indices)])
        >>> assert torch.all(alt == res)

    Ignore:
        >>> # xdoctest: +REQUIRES(module:torch, module:onnx, module:onnx_tf)
        >>> # Test if this converts to ONNX
        >>> from kwarray.util_torch import *  # NOQA
        >>> import torch.onnx
        >>> import io
        >>> import onnx
        >>> import onnx_tf.backend
        >>> import numpy as np
        >>> data = torch.from_numpy(np.array([
        >>>     [0, 1, 2],
        >>>     [3, 4, 5],
        >>>     [6, 7, 8],
        >>>     [9, 10, 11],
        >>> ]))
        >>> indices = torch.from_numpy(np.array([0, 1, 2, 1])).long()
        >>> class TFConvertWrapper(torch.nn.Module):
        >>>     def forward(self, data, indices):
        >>>         return one_hot_lookup(data, indices)
        >>> ###
        >>> # Test the ONNX export
        >>> wrapped = TFConvertWrapper()
        >>> onnx_file = io.BytesIO()
        >>> torch.onnx.export(
        >>>     wrapped, tuple([data, indices]),
        >>>     input_names=['data', 'indices'],
        >>>     output_names=['out'],
        >>>     f=onnx_file,
        >>>     opset_version=11,
        >>>     verbose=1,
        >>> )
        >>> onnx_file.seek(0)
        >>> onnx_model = onnx.load(onnx_file)
        >>> onnx_tf_model = onnx_tf.backend.prepare(onnx_model)
        >>> # Test that the resulting graph tensors are concretely sized.
        >>> import tensorflow as tf
        >>> onnx_gd = onnx_tf_model.graph.as_graph_def()
        >>> output_tensors = tf.import_graph_def(
        >>>     onnx_gd,
        >>>     input_map={},
        >>>     return_elements=[onnx_tf_model.tensor_dict[ol].name for ol in onnx_tf_model.outputs]
        >>> )
        >>> assert all(isinstance(d.value, int) for t in output_tensors for d in t.shape)
        >>> tf_outputs = onnx_tf_model.run([data, indices])
        >>> pt_outputs = wrapped(data, indices)
        >>> print('tf_outputs = {!r}'.format(tf_outputs))
        >>> print('pt_outputs = {!r}'.format(pt_outputs))
        >>> ###
        >>> # Test if data is more than 2D
        >>> shape = (4, 3, 8)
        >>> data = torch.arange(int(np.prod(shape))).view(*shape).float()
        >>> indices = torch.from_numpy(np.array([0, 1, 2, 1])).long()
        >>> onnx_file = io.BytesIO()
        >>> torch.onnx.export(
        >>>     wrapped, tuple([data, indices]),
        >>>     input_names=['data', 'indices'],
        >>>     output_names=['out'],
        >>>     f=onnx_file,
        >>>     opset_version=11,
        >>>     verbose=1,
        >>> )
        >>> onnx_file.seek(0)
        >>> onnx_model = onnx.load(onnx_file)
        >>> onnx_tf_model = onnx_tf.backend.prepare(onnx_model)
        >>> # Test that the resulting graph tensors are concretely sized.
        >>> import tensorflow as tf
        >>> onnx_gd = onnx_tf_model.graph.as_graph_def()
        >>> output_tensors = tf.import_graph_def(
        >>>     onnx_gd,
        >>>     input_map={},
        >>>     return_elements=[onnx_tf_model.tensor_dict[ol].name for ol in onnx_tf_model.outputs]
        >>> )
        >>> assert all(isinstance(d.value, int) for t in output_tensors for d in t.shape)
        >>> tf_outputs = onnx_tf_model.run([data, indices])
        >>> pt_outputs = wrapped(data, indices)
        >>> print('tf_outputs = {!r}'.format(tf_outputs))
        >>> print('pt_outputs = {!r}'.format(pt_outputs))
    """
    if torch is not None and torch.is_tensor(indices):
        if torch.onnx.is_in_onnx_export():

            # Don't use eye for ONNX
            ASSUME_OPTSET = 10

            device = indices.device

            n = data.shape[1]
            # Have to construct eye manually to satisfy onnx
            # Manually construct diag indices
            row_idxs = torch.arange(n, device=device)
            eye_idxs = row_idxs + (row_idxs * n)

            if ASSUME_OPTSET >= 11:
                # With opset 11 we use the "put" operation to directly
                # populate the diagonal elements.
                eye = torch.zeros((n, n), dtype=data.dtype, device=device)
                flat_eye = eye.view(n * n)
                diag_elem = torch.ones(n, dtype=data.dtype, device=device)
                flat_eye[eye_idxs] = diag_elem
            elif ASSUME_OPTSET >= 10:
                # With opset 10 we cannot use "put", so we have to get spicey
                # Construct the flat indexes of an NxN matrix
                flat_idxs = torch.arange(n * n)
                # Broadcast and check if these flat indexes are equal to the
                # target indexes, then sum over the broadcast dimension
                flat_eye = (flat_idxs[:, None] == eye_idxs[None, :]).to(data.dtype).sum(dim=1)
            else:
                raise AssertionError('ASSUME_OPTSET = {}'.format(ASSUME_OPTSET))

            eye = flat_eye.view(n, n)

            # Do the normal lookup in the eye matrix to get the OHE
            ohe = eye[indices]
            # need to pad OHE with extra dimensions for broadcasting
            extra_dims = len(data.shape) - 2
            if extra_dims > 0:
                ohe = ohe[(Ellipsis,) + (None,) * extra_dims]

            # Have to use multiply trick to satisfy onnx
            out = (data * ohe).sum(dim=1)
        else:
            ohe = torch.eye(data.shape[1], dtype=torch.bool, device=indices.device)[indices]
            out = data[ohe]
    else:
        # ohe = kwarray.one_hot_embedding(indices, data.shape[1]).astype(np.bool)
        # Constructing the OHE with a small dtype offers a sizable speed advantage
        ohe = np.eye(data.shape[1], dtype=np.bool)[indices]
        out = data[ohe]
    return out


if __name__ == '__main__':
    """
    CommandLine:
        python -m kwarray.util_torch all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
"""
Handle and interchange between different random number generators (numpy,
python, torch, ...). Also defines useful random iterator functions and
:func:`ensure_rng`.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import random
import itertools as it

_SEED_MAX = int(2 ** 32 - 1)


def seed_global(seed, offset=0):
    """
    Seeds the python, numpy, and torch global random states

    Args:
        seed (int): seed to use
        offset (int, optional): if specified, uses a different seed for each
            global random state separated by this offset.
    """
    random.seed((seed) % _SEED_MAX)
    np.random.seed((seed + offset) % _SEED_MAX)
    try:
        import torch
    except ImportError:
        pass
    else:
        torch.random.manual_seed((seed + 2 * offset) % _SEED_MAX)
        torch.cuda.manual_seed_all((seed + 3 * offset) % _SEED_MAX)


def shuffle(items, rng=None):
    """
    Shuffles a list inplace and then returns it for convinience

    Args:
        items (list or ndarray): list to shuffle
        rng (RandomState or int): seed or random number gen

    Returns:
        list: this is the input, but returned for convinience

    Example:
        >>> list1 = [1, 2, 3, 4, 5, 6]
        >>> list2 = shuffle(list(list1), rng=1)
        >>> assert list1 != list2
        >>> result = str(list2)
        >>> print(result)
        [3, 2, 5, 1, 4, 6]
    """
    rng = ensure_rng(rng)
    rng.shuffle(items)
    return items


def random_combinations(items, size, num=None, rng=None):
    """
    Yields ``num`` combinations of length ``size`` from items in random order

    Args:
        items (List): pool of items to choose from
        size (int): number of items in each combination
        num (None, default=None): number of combinations to generate
        rng (int | RandomState, default=None): seed or random number generator

    Yields:
        Tuple: a random combination of ``items`` of length ``size``.

    Example:
        >>> import ubelt as ub
        >>> items = list(range(10))
        >>> size = 3
        >>> num = 5
        >>> rng = 0
        >>> # xdoctest: +IGNORE_WANT
        >>> combos = list(random_combinations(items, size, num, rng))
        >>> print('combos = {}'.format(ub.repr2(combos, nl=1)))
        combos = [
            (0, 6, 9),
            (4, 7, 8),
            (4, 6, 7),
            (2, 3, 5),
            (1, 2, 4),
        ]

    Example:
        >>> import ubelt as ub
        >>> items = list(zip(range(10), range(10)))
        >>> # xdoctest: +IGNORE_WANT
        >>> combos = list(random_combinations(items, 3, num=5, rng=0))
        >>> print('combos = {}'.format(ub.repr2(combos, nl=1)))
        combos = [
            ((0, 0), (6, 6), (9, 9)),
            ((4, 4), (7, 7), (8, 8)),
            ((4, 4), (6, 6), (7, 7)),
            ((2, 2), (3, 3), (5, 5)),
            ((1, 1), (2, 2), (4, 4)),
        ]

    """
    import scipy.special
    rng = ensure_rng(rng, api='python')
    num_ = np.inf if num is None else num
    # Ensure we dont request more than is possible
    n_max = int(scipy.special.comb(len(items), size))
    num_ = min(n_max, num_)
    if num is not None and num_ > n_max // 2:
        # If num is too big just generate all combinations and shuffle them
        combos = list(it.combinations(items, size))
        rng.shuffle(combos)
        for combo in combos[:num]:
            yield combo
    else:
        # Otherwise yield randomly until we get something we havent seen
        items = list(items)
        combos = set()
        while len(combos) < num_:
            # combo = tuple(sorted(rng.choice(items, size, replace=False)))
            combo = tuple(sorted(rng.sample(items, size)))
            if combo not in combos:
                # TODO: store indices instead of combo values
                combos.add(combo)
                yield combo


def random_product(items, num=None, rng=None):
    """
    Yields ``num`` items from the cartesian product of items in a random order.

    Args:
        items (List[Sequence]):
            items to get caresian product of packed in a list or tuple.
            (note this deviates from api of :func:`itertools.product`)

        num (int, default=None):
            maximum number of items to generate. If None, all

        rng (random.Random | np.random.RandomState | int):
            random number generator

    Yields:
        Tuple: a random item in the cartesian product

    Example:
        >>> import ubelt as ub
        >>> items = [(1, 2, 3), (4, 5, 6, 7)]
        >>> rng = 0
        >>> # xdoctest: +IGNORE_WANT
        >>> products = list(random_product(items, rng=0))
        >>> print(ub.repr2(products, nl=0))
        [(3, 4), (1, 7), (3, 6), (2, 7),... (1, 6), (2, 5), (2, 4)]
        >>> products = list(random_product(items, num=3, rng=0))
        >>> print(ub.repr2(products, nl=0))
        [(3, 4), (1, 7), (3, 6)]

    Example:
        >>> # xdoctest: +REQUIRES(--profile)
        >>> rng = ensure_rng(0)
        >>> items = [np.array([15, 14]), np.array([27, 26]),
        >>>          np.array([21, 22]), np.array([32, 31])]
        >>> num = 2
        >>> for _ in range(100):
        >>>     list(random_product(items, num=num, rng=rng))
    """
    # NUMPY_RNG = True  # toggle new speedup on
    try:
        if not isinstance(items, (list, tuple)):
            raise TypeError
        idx_cards = np.array([len(g) for g in items], dtype=np.uint32)
    except (TypeError, AttributeError):
        items = [list(g) for g in items]
        idx_cards = np.array([len(g) for g in items], dtype=np.uint32)

    ndims = len(items)
    # max_num = np.prod(idx_cards.astype(np.float))
    max_num = np.multiply.reduce(idx_cards.astype(np.float32))

    if num is None:
        num = max_num
    else:
        num = min(num, max_num)
        # if num > max_num:
        #     raise ValueError('num exceedes maximum number of products')

    # TODO: make this more efficient when num is large
    if max_num > 100 and num > max_num // 2:

        rng = ensure_rng(rng, 'python')

        for prod in shuffle(list(it.product(*items)), rng=rng):
            yield prod
    else:
        if True:  # NUMPY_RNG
            rng = ensure_rng(rng, 'numpy')
            # Need to use least-common-multiple so the mod of all idxs
            # are equally likely
            card_lcm = np.lcm.reduce(idx_cards)
        else:
            rng = ensure_rng(rng, 'python')

        seen = set()
        while len(seen) < num:

            if True:  # NUMPY_RNG
                idxs = rng.randint(0, card_lcm, size=ndims, dtype=idx_cards.dtype)
                idxs %= idx_cards
                idxs = tuple(idxs.tolist())
            else:
                idxs = tuple(rng.randint(0, n - 1) for n in idx_cards)

            if idxs not in seen:
                seen.add(idxs)
                prod = tuple(g[x] for g, x in zip(items, idxs))
                yield prod


def _npstate_to_pystate(npstate):
    """
    Convert state of a NumPy RandomState object to a state
    that can be used by Python's Random. Derived from [1]_.

    References:
        .. [1] https://stackoverflow.com/questions/44313620/convert-randomstate

    Example:
        >>> py_rng = random.Random(0)
        >>> np_rng = np.random.RandomState(seed=0)
        >>> npstate = np_rng.get_state()
        >>> pystate = _npstate_to_pystate(npstate)
        >>> py_rng.setstate(pystate)
        >>> assert np_rng.rand() == py_rng.random()
    """
    PY_VERSION = 3
    version, keys, pos, has_gauss, cached_gaussian_ = npstate
    keys_pos = tuple(map(int, keys)) + (int(pos),)
    cached_gaussian_ = cached_gaussian_ if has_gauss else None
    pystate = (PY_VERSION, keys_pos, cached_gaussian_)
    return pystate


def _pystate_to_npstate(pystate):
    """
    Convert state of a Python Random object to state usable
    by NumPy RandomState. Derived from [2]_.

    References:
        .. [2] https://stackoverflow.com/questions/44313620/convert-randomstate

    Example:
        >>> py_rng = random.Random(0)
        >>> np_rng = np.random.RandomState(seed=0)
        >>> pystate = py_rng.getstate()
        >>> npstate = _pystate_to_npstate(pystate)
        >>> np_rng.set_state(npstate)
        >>> assert np_rng.rand() == py_rng.random()
    """
    NP_VERSION = 'MT19937'
    version, keys_pos_, cached_gaussian_ = pystate
    keys, pos = keys_pos_[:-1], keys_pos_[-1]
    keys = np.array(keys, dtype=np.uint32)
    has_gauss = cached_gaussian_ is not None
    cached_gaussian = cached_gaussian_ if has_gauss else 0.0
    npstate = (NP_VERSION, keys, pos, has_gauss, cached_gaussian)
    return npstate


def _coerce_rng_type(rng):
    """
    Internal method that transforms input seeds into an integer form.
    """
    if rng is None or isinstance(rng, (random.Random, np.random.RandomState)):
        pass
    elif rng is random:
        rng = rng._inst
    elif rng is np.random:
        rng = np.random.mtrand._rand
    elif isinstance(rng, (float, np.floating)):
        rng = float(rng)
        # Coerce the float into an integer
        a, b = rng.as_integer_ratio()
        if b == 1:
            rng = a
        else:
            s = max(a.bit_length(), b.bit_length())
            rng = (b << s) | a
    elif isinstance(rng, (int, np.integer)):
        rng = int(rng)
    else:
        raise TypeError(
            'Cannot coerce {!r} to a random object'.format(type(rng)))
    return rng


def ensure_rng(rng, api='numpy'):
    """
    Coerces input into a random number generator.

    This function is useful for ensuring that your code uses a controlled
    internal random state that is independent of other modules.

    If the input is None, then a global random state is returned.

    If the input is a numeric value, then that is used as a seed to construct a
    random state.

    If the input is a random number generator, then another random number
    generator with the same state is returned. Depending on the api, this
    random state is either return as-is, or used to construct an equivalent
    random state with the requested api.

    Args:
        rng (int | float | numpy.random.RandomState | random.Random | None):
            if None, then defaults to the global rng. Otherwise this can
            be an integer or a RandomState class

        api (str, default='numpy'): specify the type of random number
            generator to use. This can either be 'numpy' for a
            :class:`numpy.random.RandomState` object or 'python' for a
            :class:`random.Random` object.

    Returns:
        (numpy.random.RandomState | random.Random) : rng -
            either a numpy or python random number generator, depending on the
            setting of ``api``.

    Example:
        >>> rng = ensure_rng(None)
        >>> ensure_rng(0).randint(0, 1000)
        684
        >>> ensure_rng(np.random.RandomState(1)).randint(0, 1000)
        37

    Example:
        >>> num = 4
        >>> print('--- Python as PYTHON ---')
        >>> py_rng = random.Random(0)
        >>> pp_nums = [py_rng.random() for _ in range(num)]
        >>> print(pp_nums)
        >>> print('--- Numpy as PYTHON ---')
        >>> np_rng = ensure_rng(random.Random(0), api='numpy')
        >>> np_nums = [np_rng.rand() for _ in range(num)]
        >>> print(np_nums)
        >>> print('--- Numpy as NUMPY---')
        >>> np_rng = np.random.RandomState(seed=0)
        >>> nn_nums = [np_rng.rand() for _ in range(num)]
        >>> print(nn_nums)
        >>> print('--- Python as NUMPY---')
        >>> py_rng = ensure_rng(np.random.RandomState(seed=0), api='python')
        >>> pn_nums = [py_rng.random() for _ in range(num)]
        >>> print(pn_nums)
        >>> assert np_nums == pp_nums
        >>> assert pn_nums == nn_nums

    Example:
        >>> # Test that random modules can be coerced
        >>> import random
        >>> import numpy as np
        >>> ensure_rng(random, api='python')
        >>> ensure_rng(random, api='numpy')
        >>> ensure_rng(np.random, api='python')
        >>> ensure_rng(np.random, api='numpy')

    Ignore:
        >>> np.random.seed(0)
        >>> np.random.randint(0, 10000)
        2732
        >>> np.random.seed(0)
        >>> np.random.mtrand._rand.randint(0, 10000)
        2732
        >>> np.random.seed(0)
        >>> ensure_rng(None).randint(0, 10000)
        2732
        >>> np.random.randint(0, 10000)
        9845
        >>> ensure_rng(None).randint(0, 10000)
        3264
    """
    rng = _coerce_rng_type(rng)

    if api == 'numpy':
        if rng is None:
            # This is the underlying random state of the np.random module
            rng = np.random.mtrand._rand
            # Dont do this because it seeds using dev/urandom
            # rng = np.random.RandomState(seed=None)
        elif isinstance(rng, int):
            rng = np.random.RandomState(seed=rng % _SEED_MAX)
        elif isinstance(rng, random.Random):
            # Convert python to numpy random state
            py_rng = rng
            pystate = py_rng.getstate()
            npstate = _pystate_to_npstate(pystate)
            rng = np_rng = np.random.RandomState(seed=0)
            np_rng.set_state(npstate)
    elif api == 'python':
        if rng is None:
            # This is the underlying random state of the random module
            rng = random._inst
        elif isinstance(rng, int):
            rng = random.Random(rng % _SEED_MAX)
        elif isinstance(rng, np.random.RandomState):
            # Convert numpy to python random state
            np_rng = rng
            npstate = np_rng.get_state()
            pystate = _npstate_to_pystate(npstate)
            rng = py_rng = random.Random(0)
            py_rng.setstate(pystate)
    else:
        raise KeyError('unknown rng api={}'.format(api))
    return rng


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m kwarray.util_random
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -* -coding: utf-8 -*-
"""
Fast 32-bit random functions for numpy as of 2018. (More recent versions of
numpy may have these natively supported).
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np


def uniform(low=0.0, high=1.0, size=None, dtype=np.float32, rng=np.random):
    """
    Draws float32 samples from a uniform distribution.

    Samples are uniformly distributed over the half-open interval
    ``[low, high)`` (includes low, but excludes high).

    Args:
        low (float, default=0.0):
            Lower boundary of the output interval.  All values generated will
            be greater than or equal to low.

        high (float, default=1.0):
            Upper boundary of the output interval.  All values generated will
            be less than high.

        size (int | Tuple[int], default=None):
            Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  If size is ``None`` (default),
            a single value is returned if ``low`` and ``high`` are both scalars.
            Otherwise, ``np.broadcast(low, high).size`` samples are drawn.

        dtype (type): either np.float32 or np.float64

        rng (numpy.random.RandomState): underlying random state

    Returns:
        ndarray[dtype]: normally distributed random numbers with chosen dtype

    Benchmark:
        >>> from timerit import Timerit
        >>> import kwarray
        >>> size = (300, 300, 3)
        >>> for timer in Timerit(100, bestof=10, label='dtype=np.float32'):
        >>>     rng = kwarray.ensure_rng(0)
        >>>     with timer:
        >>>         ours = standard_normal(size, rng=rng, dtype=np.float32)
        >>> # Timed best=4.705 ms, mean=4.75 ± 0.085 ms for dtype=np.float32
        >>> for timer in Timerit(100, bestof=10, label='dtype=np.float64'):
        >>>     rng = kwarray.ensure_rng(0)
        >>>     with timer:
        >>>         theirs = standard_normal(size, rng=rng, dtype=np.float64)
        >>> # Timed best=9.327 ms, mean=9.794 ± 0.4 ms for rng.np.float64
    """
    if dtype is np.float32:
        return uniform32(low, high, size, rng)
    elif dtype is np.float64:
        return rng.uniform(low, high, size)
    else:
        raise ValueError('dtype = {!r}'.format(dtype))


def standard_normal(size, mean=0, std=1, dtype=float, rng=np.random):
    """
    Draw samples from a standard Normal distribution with a specified mean and
    standard deviation.

    Args:
        size (int | Tuple[int, *int]) : shape of the returned ndarray
        mean (float, default=0): mean of the normal distribution
        std (float, default=1): standard deviation of the normal distribution
        dtype (type): either np.float32 or np.float64
        rng (numpy.random.RandomState): underlying random state

    Returns:
        ndarray[dtype]: normally distributed random numbers with chosen dtype

    Benchmark:
        >>> from timerit import Timerit
        >>> import kwarray
        >>> size = (300, 300, 3)
        >>> for timer in Timerit(100, bestof=10, label='dtype=np.float32'):
        >>>     rng = kwarray.ensure_rng(0)
        >>>     with timer:
        >>>         ours = standard_normal(size, rng=rng, dtype=np.float32)
        >>> # Timed best=4.705 ms, mean=4.75 ± 0.085 ms for dtype=np.float32
        >>> for timer in Timerit(100, bestof=10, label='dtype=np.float64'):
        >>>     rng = kwarray.ensure_rng(0)
        >>>     with timer:
        >>>         theirs = standard_normal(size, rng=rng, dtype=np.float64)
        >>> # Timed best=9.327 ms, mean=9.794 ± 0.4 ms for rng.np.float64
    """
    if dtype is np.float32:
        return standard_normal32(size, mean, std, rng)
    elif dtype is np.float64:
        return standard_normal64(size, mean, std, rng)
    else:
        raise ValueError('dtype = {!r}'.format(dtype))


def standard_normal32(size, mean=0, std=1, rng=np.random):
    """
    Fast normally distributed random variables using the Box–Muller transform

    The difference between this function and
    :func:`numpy.random.standard_normal` is that we use float32 arrays in the
    backend instead of float64.  Halving the amount of bits that need to be
    manipulated can significantly reduce the execution time, and 32-bit
    precision is often good enough.

    Args:
        size (int | Tuple[int, *int]) : shape of the returned ndarray
        mean (float, default=0): mean of the normal distribution
        std (float, default=1): standard deviation of the normal distribution
        rng (numpy.random.RandomState): underlying random state

    Returns:
        ndarray[float32]: normally distributed random numbers

    References:
        https://en.wikipedia.org/wiki/Box%E2%80%93Muller_transform

    SeeAlso:
        * standard_normal
        * standard_normal64

    Example:
        >>> import scipy
        >>> import scipy.stats
        >>> pts = 1000
        >>> # Our numbers are normally distributed with high probability
        >>> rng = np.random.RandomState(28041990)
        >>> ours_a = standard_normal32(pts, rng=rng)
        >>> ours_b = standard_normal32(pts, rng=rng) + 2
        >>> ours = np.concatenate((ours_a, ours_b))  # numerical stability?
        >>> p = scipy.stats.normaltest(ours)[1]
        >>> print('Probability our data is non-normal is: {:.4g}'.format(p))
        Probability our data is non-normal is: 1.573e-14
        >>> rng = np.random.RandomState(28041990)
        >>> theirs_a = rng.standard_normal(pts)
        >>> theirs_b = rng.standard_normal(pts) + 2
        >>> theirs = np.concatenate((theirs_a, theirs_b))
        >>> p = scipy.stats.normaltest(theirs)[1]
        >>> print('Probability their data is non-normal is: {:.4g}'.format(p))
        Probability their data is non-normal is: 3.272e-11

    Example:
        >>> pts = 1000
        >>> rng = np.random.RandomState(28041990)
        >>> ours = standard_normal32(pts, mean=10, std=3, rng=rng)
        >>> assert np.abs(ours.std() - 3.0) < 0.1
        >>> assert np.abs(ours.mean() - 10.0) < 0.1

    Example:
        >>> # Test an even and odd numbers of points
        >>> assert standard_normal32(3).shape == (3,)
        >>> assert standard_normal32(2).shape == (2,)
        >>> assert standard_normal32(1).shape == (1,)
        >>> assert standard_normal32(0).shape == (0,)
        >>> assert standard_normal32((3, 1)).shape == (3, 1)
        >>> assert standard_normal32((3, 0)).shape == (3, 0)
    """
    # return rng.standard_normal(size) * std + mean
    # The integer and float dtype must have the same number of bits
    dtype = np.float32
    int_dtype = np.uint32
    MAX_INT = np.core.getlimits.iinfo(int_dtype).max

    # Preallocate output
    out = np.empty(size, dtype=dtype)
    total = out.size
    n = int(np.ceil(total / 2))

    # Generate uniform-01 random numbers that we will transform
    u12 = rng.randint(1, MAX_INT - 1, size=n * 2, dtype=int_dtype).astype(dtype)
    np.divide(u12, dtype(MAX_INT), out=u12)

    u1 = u12[:n]
    u2 = u12[n:]
    # Box–Muller transform transforms two uniformly distributed values
    # into normally distributed random values. Because of the way it works
    # we generate two sets of uniform random values, and then
    flat_out = out.ravel()

    # The following logic is equivalent to:
    #   R = std * np.sqrt(-2 * np.log(u1))
    #   Theta = 2 * np.pi * u2
    #   flat_out[0:n] = R * np.cos(Theta) + mean
    #   flat_out[-n:] = R * np.sin(Theta) + mean
    np.log(u1, out=u1)
    np.multiply(-2 * (std ** 2), u1, out=u1)
    R = np.sqrt(u1, out=u1)

    Theta = np.multiply(2 * np.pi, u2, out=u2)

    out1 = flat_out[0:n]
    out2 = flat_out[-n:]
    np.cos(Theta, out=out1)
    np.sin(Theta, out=out2)

    np.multiply(R, out1, out=out1)
    np.multiply(R, out2, out=out2)

    if mean != 0:
        np.add(out, mean, out=out)
    return out


def standard_normal64(size, mean=0, std=1, rng=np.random):
    """
    Simple wrapper around rng.standard_normal to make an API compatible with
    :func:`standard_normal32`.

    Args:
        size (int | Tuple[int, *int]) : shape of the returned ndarray
        mean (float, default=0): mean of the normal distribution
        std (float, default=1): standard deviation of the normal distribution
        rng (numpy.random.RandomState): underlying random state

    Returns:
        ndarray[float64]: normally distributed random numbers

    SeeAlso:
        * standard_normal
        * standard_normal32

    Example:
        >>> pts = 1000
        >>> rng = np.random.RandomState(28041994)
        >>> out = standard_normal64(pts, mean=10, std=3, rng=rng)
        >>> assert np.abs(out.std() - 3.0) < 0.1
        >>> assert np.abs(out.mean() - 10.0) < 0.1
    """
    out = rng.standard_normal(size)
    if std != 1:
        np.multiply(out, std, out=out)
    if mean != 0:
        np.add(out, mean, out=out)
    return out


def uniform32(low=0.0, high=1.0, size=None, rng=np.random):
    """
    Draws float32 samples from a uniform distribution.

    Samples are uniformly distributed over the half-open interval
    ``[low, high)`` (includes low, but excludes high).

    Args:
        low (float, default=0.0):
            Lower boundary of the output interval.  All values generated will
            be greater than or equal to low.

        high (float, default=1.0):
            Upper boundary of the output interval.  All values generated will
            be less than high.

        size (int | Tuple[int], default=None):
            Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  If size is ``None`` (default),
            a single value is returned if ``low`` and ``high`` are both scalars.
            Otherwise, ``np.broadcast(low, high).size`` samples are drawn.

    Example:
        >>> rng = np.random.RandomState(0)
        >>> uniform32(low=0.0, high=1.0, size=None, rng=rng)
        0.5488...
        >>> uniform32(low=0.0, high=1.0, size=2000, rng=rng).sum()
        1004.94...
        >>> uniform32(low=-10, high=10.0, size=2000, rng=rng).sum()
        202.44...

    Benchmark:
        >>> from timerit import Timerit
        >>> import kwarray
        >>> size = 512 * 512
        >>> for timer in Timerit(100, bestof=10, label='theirs: dtype=np.float64'):
        >>>     rng = kwarray.ensure_rng(0)
        >>>     with timer:
        >>>         theirs = rng.uniform(size=size)
        >>> for timer in Timerit(100, bestof=10, label='theirs: dtype=np.float32'):
        >>>     rng = kwarray.ensure_rng(0)
        >>>     with timer:
        >>>         theirs = rng.rand(size).astype(np.float32)
        >>> for timer in Timerit(100, bestof=10, label='ours: dtype=np.float32'):
        >>>     rng = kwarray.ensure_rng(0)
        >>>     with timer:
        >>>         ours = uniform32(size=size)
    """
    if size is None:
        out = np.float32(rng.uniform(low, high))
    else:
        total = size if isinstance(size, int) else np.prod(size)
        if total < 1764:
            # For smaller sizes simply casting to float32 is faster
            out = rng.uniform(low, high, size).astype(np.float32)
        else:
            # This is only faster for large sizes.
            # The cuttoff size seems to be (42 * 42).
            dtype = np.float32
            int_dtype = np.uint32
            MAX_INT = np.core.getlimits.iinfo(int_dtype).max
            # Generate uniform-01 random numbers that we will transform
            out = rng.randint(0, MAX_INT - 1, size=size, dtype=int_dtype).astype(dtype)
            if low == 0 and high == 1.0:
                np.divide(out, dtype(MAX_INT), out=out)
            else:
                extent = dtype(MAX_INT) / (high - low)
                np.divide(out, extent, out=out)
                np.add(out, low, out=out)
    return out



# -*- coding: utf-8 -*-
"""
Functions for partitioning numpy arrays into groups.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import ubelt as ub


def group_items(item_list, groupid_list, assume_sorted=False, axis=None):
    """
    Groups a list of items by group id.

    Works like :func:`ubelt.group_items`, but with numpy optimizations.
    This can be quite a bit faster than using :func:`itertools.groupby` [1]_
    [2]_.

    In cases where there are many lists of items to group (think column-major
    data), consider using :func:`group_indices` and :func:`apply_grouping`
    instead.

    Args:
        item_list (ndarray[T1]):
            The input array of items to group.

        groupid_list (ndarray[T2]):
            Each item is an id corresponding to the item at the same position
            in ``item_list``.  For the fastest runtime, the input array must be
            numeric (ideally with integer types). This list must be
            1-dimensional.

        assume_sorted (bool, default=False):
            If the input array is sorted, then setting this to True will avoid
            an unnecessary sorting operation and improve efficiency.

        axis (int | None):
            group along a particular axis in ``items`` if it is n-dimensional

    Returns:
        Dict[T2, ndarray[T1]]: mapping from groupids to corresponding items

    References:
        .. [1] http://stackoverflow.com/questions/4651683/
        .. [2] numpy-grouping-using-itertools-groupby-performance

    Example:
        >>> from kwarray.util_groups import *  # NOQA
        >>> items = np.array([0, 1, 2, 3, 4, 5, 6, 7])
        >>> keys = np.array( [2, 2, 1, 1, 0, 1, 0, 1])
        >>> grouped = group_items(items, keys)
        >>> print(ub.repr2(grouped, nl=1, with_dtype=False))
        {
            0: np.array([4, 6]),
            1: np.array([2, 3, 5, 7]),
            2: np.array([0, 1]),
        }
    """
    keys, groupxs = group_indices(groupid_list, assume_sorted=assume_sorted)
    grouped_values = apply_grouping(item_list, groupxs, axis=axis)
    grouped = dict(zip(keys, grouped_values))
    return grouped


def group_indices(idx_to_groupid, assume_sorted=False):
    """
    Find unique items and the indices at which they appear in an array.

    A common use case of this function is when you have a list of objects
    (often numeric but sometimes not) and an array of "group-ids" corresponding
    to that list of objects.

    Using this function will return a list of indices that can be used in
    conjunction with :func:`apply_grouping` to group the elements.  This is
    most useful when you have many lists (think column-major data)
    corresponding to the group-ids.

    In cases where there is only one list of objects or knowing the indices
    doesn't matter, then consider using func:`group_items` instead.

    Args:
        idx_to_groupid (ndarray):
            The input array, where each item is interpreted as a group id.
            For the fastest runtime, the input array must be numeric (ideally
            with integer types).  If the type is non-numeric then the less
            efficient :func:`ubelt.group_items` is used.

        assume_sorted (bool, default=False):
            If the input array is sorted, then setting this to True will avoid
            an unnecessary sorting operation and improve efficiency.

    Returns:
        Tuple[ndarray, List[ndarrays]]: (keys, groupxs) -
            keys (ndarray):
                The unique elements of the input array in order
            groupxs (List[ndarray]):
                Corresponding list of indexes.  The i-th item is an array
                indicating the indices where the item ``key[i]`` appeared in
                the input array.

    Example:
        >>> # xdoctest: +IGNORE_WHITESPACE
        >>> import ubelt as ub
        >>> idx_to_groupid = np.array([2, 1, 2, 1, 2, 1, 2, 3, 3, 3, 3])
        >>> (keys, groupxs) = group_indices(idx_to_groupid)
        >>> print(ub.repr2(keys, with_dtype=False))
        >>> print(ub.repr2(groupxs, with_dtype=False))
        np.array([1, 2, 3])
        [
            np.array([1, 3, 5]),
            np.array([0, 2, 4, 6]),
            np.array([ 7,  8,  9, 10]),
        ]

    Example:
        >>> # xdoctest: +IGNORE_WHITESPACE
        >>> import ubelt as ub
        >>> idx_to_groupid = np.array([[  24], [ 129], [ 659], [ 659], [ 24],
        ...       [659], [ 659], [ 822], [ 659], [ 659], [24]])
        >>> # 2d arrays must be flattened before coming into this function so
        >>> # information is on the last axis
        >>> (keys, groupxs) = group_indices(idx_to_groupid.T[0])
        >>> print(ub.repr2(keys, with_dtype=False))
        >>> print(ub.repr2(groupxs, with_dtype=False))
        np.array([ 24, 129, 659, 822])
        [
            np.array([ 0,  4, 10]),
            np.array([1]),
            np.array([2, 3, 5, 6, 8, 9]),
            np.array([7]),
        ]

    Example:
        >>> # xdoctest: +IGNORE_WHITESPACE
        >>> import ubelt as ub
        >>> idx_to_groupid = np.array([True, True, False, True, False, False, True])
        >>> (keys, groupxs) = group_indices(idx_to_groupid)
        >>> print(ub.repr2(keys, with_dtype=False))
        >>> print(ub.repr2(groupxs, with_dtype=False))
        np.array([False,  True])
        [
            np.array([2, 4, 5]),
            np.array([0, 1, 3, 6]),
        ]

    Example:
        >>> # xdoctest: +IGNORE_WHITESPACE
        >>> import ubelt as ub
        >>> idx_to_groupid = [('a', 'b'),  ('d', 'b'), ('a', 'b'), ('a', 'b')]
        >>> (keys, groupxs) = group_indices(idx_to_groupid)
        >>> print(ub.repr2(keys, with_dtype=False))
        >>> print(ub.repr2(groupxs, with_dtype=False))
        [
            ('a', 'b'),
            ('d', 'b'),
        ]
        [
            np.array([0, 2, 3]),
            np.array([1]),
        ]
    """
    _idx_to_groupid_orig = idx_to_groupid
    idx_to_groupid = np.array(idx_to_groupid, copy=False)
    _n_item = idx_to_groupid.size
    _dtype = idx_to_groupid.dtype
    _kind = _dtype.kind
    if _kind == 'U' or _kind == 'O':
        # fallback to slower algorithm for non-numeric data
        group = ub.group_items(range(_n_item), _idx_to_groupid_orig)
        try:
            # attempt to return values in a consistant order
            sortx = ub.argsort(list(group.keys()))
            keys = list(ub.take(list(group.keys()), sortx))
            groupxs = list(ub.take(list(map(np.array, group.values())), sortx))
        except Exception:
            keys = list(group.keys())
            groupxs = list(map(np.array, group.values()))
        return keys, groupxs

    # Sort items and idx_to_groupid by groupid
    if assume_sorted:
        sortx = np.arange(len(idx_to_groupid))
        groupids_sorted = idx_to_groupid
    else:
        sortx = idx_to_groupid.argsort()
        groupids_sorted = idx_to_groupid.take(sortx)

    if _kind == 'b':
        # Ensure bools are internally cast to integers
        # However, be sure that the groups are returned as the original dtype
        _groupids = groupids_sorted.astype(np.int8)
    else:
        _groupids = groupids_sorted

    # Find the boundaries between groups
    diff = np.ones(_n_item + 1, _groupids.dtype)
    np.subtract(_groupids[1:], _groupids[:-1], out=diff[1:_n_item])
    idxs = np.flatnonzero(diff)
    # Groups are between bounding indexes
    groupxs = [sortx[lx:rx] for lx, rx in zip(idxs, idxs[1:])]  # 34.5%
    # Unique group keys
    keys = groupids_sorted[idxs[:-1]]
    return keys, groupxs


def apply_grouping(items, groupxs, axis=0):
    """
    Applies grouping from group_indicies.

    Typically used in conjunction with :func:`group_indices`.

    Args:
        items (ndarray): items to group

        groupxs (List[ndarrays[int]]): groups of indices

        axis (None|int, default=0) axis along which to group

    Returns:
        List[ndarray]: grouped items


    Example:
        >>> # xdoctest: +IGNORE_WHITESPACE
        >>> idx_to_groupid = np.array([2, 1, 2, 1, 2, 1, 2, 3, 3, 3, 3])
        >>> items          = np.array([1, 8, 5, 5, 8, 6, 7, 5, 3, 0, 9])
        >>> (keys, groupxs) = group_indices(idx_to_groupid)
        >>> grouped_items = apply_grouping(items, groupxs)
        >>> result = str(grouped_items)
        >>> print(result)
        [array([8, 5, 6]), array([1, 5, 8, 7]), array([5, 3, 0, 9])]
    """
    # Should there be a contiguous check here?
    # items_ = np.ascontiguousarray(items)
    return [items.take(xs, axis=axis) for xs in groupxs]


def group_consecutive(arr, offset=1):
    """
    Returns lists of consecutive values. Implementation inspired by [3]_.

    Args:
        arr (ndarray):
            array of ordered values

        offset (float, default=1):
            any two values separated by this offset are grouped.  In the
            default case, when offset=1, this groups increasing values like: 0,
            1, 2. When offset is 0 it groups consecutive values thta are the
            same, e.g.: 4, 4, 4.

    Returns:
        List[ndarray]: a list of arrays that are the groups from the input

    Notes:
        This is equivalent (and faster) to using:
        apply_grouping(data, group_consecutive_indices(data))

    References:
        .. [3] http://stackoverflow.com/questions/7352684/groups-consecutive-elements

    Example:
        >>> arr = np.array([1, 2, 3, 5, 6, 7, 8, 9, 10, 15, 99, 100, 101])
        >>> groups = group_consecutive(arr)
        >>> print('groups = {}'.format(list(map(list, groups))))
        groups = [[1, 2, 3], [5, 6, 7, 8, 9, 10], [15], [99, 100, 101]]
        >>> arr = np.array([0, 0, 3, 0, 0, 7, 2, 3, 4, 4, 4, 1, 1])
        >>> groups = group_consecutive(arr, offset=1)
        >>> print('groups = {}'.format(list(map(list, groups))))
        groups = [[0], [0], [3], [0], [0], [7], [2, 3, 4], [4], [4], [1], [1]]
        >>> groups = group_consecutive(arr, offset=0)
        >>> print('groups = {}'.format(list(map(list, groups))))
        groups = [[0, 0], [3], [0, 0], [7], [2], [3], [4, 4, 4], [1, 1]]
    """
    split_indicies = np.nonzero(np.diff(arr) != offset)[0] + 1
    groups = np.array_split(arr, split_indicies)
    return groups


def group_consecutive_indices(arr, offset=1):
    """
    Returns lists of indices pointing to consecutive values

    Args:
        arr (ndarray):
            array of ordered values

        offset (float, default=1):
            any two values separated by this offset are grouped.

    Returns:
        List[ndarray]: groupxs: a list of indices

    SeeAlso:

        :func:`group_consecutive`

        :func:`apply_grouping`

    Example:
        >>> arr = np.array([1, 2, 3, 5, 6, 7, 8, 9, 10, 15, 99, 100, 101])
        >>> groupxs = group_consecutive_indices(arr)
        >>> print('groupxs = {}'.format(list(map(list, groupxs))))
        groupxs = [[0, 1, 2], [3, 4, 5, 6, 7, 8], [9], [10, 11, 12]]
        >>> assert all(np.array_equal(a, b) for a, b in zip(group_consecutive(arr, 1), apply_grouping(arr, groupxs)))
        >>> arr = np.array([0, 0, 3, 0, 0, 7, 2, 3, 4, 4, 4, 1, 1])
        >>> groupxs = group_consecutive_indices(arr, offset=1)
        >>> print('groupxs = {}'.format(list(map(list, groupxs))))
        groupxs = [[0], [1], [2], [3], [4], [5], [6, 7, 8], [9], [10], [11], [12]]
        >>> assert all(np.array_equal(a, b) for a, b in zip(group_consecutive(arr, 1), apply_grouping(arr, groupxs)))
        >>> groupxs = group_consecutive_indices(arr, offset=0)
        >>> print('groupxs = {}'.format(list(map(list, groupxs))))
        groupxs = [[0, 1], [2], [3, 4], [5], [6], [7], [8, 9, 10], [11, 12]]
        >>> assert all(np.array_equal(a, b) for a, b in zip(group_consecutive(arr, 0), apply_grouping(arr, groupxs)))
    """
    split_indicies = np.nonzero(np.diff(arr) != offset)[0] + 1
    groupxs = np.array_split(np.arange(len(arr)), split_indicies)
    return groupxs


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m kwarray.util_groups
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
"""
The ``kwarray`` module implements a small set of pure-python extensions to
numpy and torch.
"""
# flake8: noqa
from __future__ import absolute_import, division, print_function, unicode_literals
"""
AutogenInit:
    mkinit ~/code/kwarray/kwarray/__init__.py --relative --nomods
    mkinit ~/code/kwarray/kwarray/__init__.py --relative --nomods -w
"""

__private__ = [
    'distributions',
    'arrayapi'
]

__explicit__ = [
    'ArrayAPI', 'dtype_info',
]

__version__ = '0.5.15'

from kwarray.arrayapi import ArrayAPI, dtype_info


# Everything after this point is autogenerated
from .algo_assignment import (maxvalue_assignment, mincost_assignment,
                              mindist_assignment,)
from .algo_setcover import (setcover,)
from .dataframe_light import (DataFrameArray, DataFrameLight, LocLight,)
from .fast_rand import (standard_normal, standard_normal32, standard_normal64,
                        uniform, uniform32,)
from .util_averages import (RunningStats, stats_dict,)
from .util_groups import (apply_grouping, group_consecutive,
                          group_consecutive_indices, group_indices,
                          group_items,)
from .util_misc import (FlatIndexer,)
from .util_numpy import (arglexmax, argmaxima, argminima, atleast_nd, boolmask,
                         isect_flags, iter_reduce_ufunc, normalize,)
from .util_random import (ensure_rng, random_combinations, random_product,
                          seed_global, shuffle,)
from .util_slider import (SlidingWindow, Stitcher,)
from .util_torch import (one_hot_embedding, one_hot_lookup,)

__all__ = ['ArrayAPI', 'DataFrameArray', 'DataFrameLight', 'FlatIndexer',
           'LocLight', 'RunningStats', 'SlidingWindow', 'Stitcher',
           'apply_grouping', 'arglexmax', 'argmaxima', 'argminima',
           'atleast_nd', 'boolmask', 'dtype_info', 'ensure_rng',
           'group_consecutive', 'group_consecutive_indices', 'group_indices',
           'group_items', 'isect_flags', 'iter_reduce_ufunc',
           'maxvalue_assignment', 'mincost_assignment', 'mindist_assignment',
           'normalize', 'one_hot_embedding', 'one_hot_lookup',
           'random_combinations', 'random_product', 'seed_global', 'setcover',
           'shuffle', 'standard_normal', 'standard_normal32',
           'standard_normal64', 'stats_dict', 'uniform', 'uniform32']
